from __future__ import division, print_function
import pyroomacoustics as pra
import numpy as np

try:
    from pyroomacoustics import build_rir

    build_rir_available = True
except:
    print("build_rir not available")
    build_rir_available = False

# tolerance for test success (1%)
tol = 0.01

fdl = 81
fs = 16000

t0 = (2 * fdl + 0.1) / fs
t1 = (3 * fdl - 0.1) / fs
t2 = (4 * fdl + 0.45) / fs
t3 = (5 * fdl + 0.001) / fs
t4 = (6 * fdl + 0.999) / fs

times = np.array(
    [
        [
            t0,
            t1 + (1 / 40 / 16000),
            t2,
        ],
        [
            t0,
            t1 + (10 / fs),
            3 * t3,
        ],
        [
            t0,
            t3,
            t4,
        ],
    ],
)
alphas = np.array(
    [
        [1.0, 0.5, -0.1],
        [0.5, 0.3, 0.1],
        [0.3, 2.0, 0.1],
    ],
)
visibilities = np.array(
    [
        [
            1,
            1,
            1,
        ],
        [
            1,
            1,
            1,
        ],
        [
            0,
            1,
            1,
        ],
    ],
    dtype=np.int32,
)


def build_rir_wrap(time, alpha, visibility, fs, fdl):

    # fractional delay length
    fdl = pra.constants.get("frac_delay_length")
    fdl2 = (fdl - 1) // 2

    # the number of samples needed
    N = int(np.ceil(time.max() * fs) + fdl)

    ir_ref = np.zeros(N)
    ir_cython = np.zeros(N)

    # Try to use the Cython extension
    build_rir.fast_rir_builder(ir_cython, time, alpha, visibility, fs, fdl)

    # fallback to pure Python implemenation
    for i in range(time.shape[0]):
        if visibility[i] == 1:
            time_ip = int(np.round(fs * time[i]))
            time_fp = (fs * time[i]) - time_ip
            ir_ref[time_ip - fdl2 : time_ip + fdl2 + 1] += alpha[
                i
            ] * pra.fractional_delay(time_fp)

    return ir_ref, ir_cython


def test_build_rir():

    if not build_rir_available:
        return

    for t, a, v in zip(times, alphas, visibilities):
        ir_ref, ir_cython = build_rir_wrap(
            times[0], alphas[0], visibilities[0], fs, fdl
        )
        assert np.max(np.abs(ir_ref - ir_cython)) < tol


def test_short():
    """Tests that an error is raised if a provided time goes below the zero index"""

    if not build_rir_available:
        return

    N = 100
    fs = 16000
    fdl = 81
    rir = np.zeros(N)

    time = np.array([0.0])
    alpha = np.array([1.0])
    visibility = np.array([1], dtype=np.int32)

    try:
        build_rir.fast_rir_builder(rir, time, alpha, visibility, fs, fdl)
        assert False
    except AssertionError:
        print("Ok, short times are caught")


def test_long():
    """Tests that an error is raised if a time falls outside the rir array"""

    if not build_rir_available:
        return

    N = 100
    fs = 16000
    fdl = 81
    rir = np.zeros(N)

    time = np.array([(N - 1) / fs])
    alpha = np.array([1.0])
    visibility = np.array([1], dtype=np.int32)

    try:
        build_rir.fast_rir_builder(rir, time, alpha, visibility, fs, fdl)
        assert False
    except AssertionError:
        print("Ok, long times are caught")


def test_errors():
    """Tests that errors are raised when array lengths differ"""

    if not build_rir_available:
        return

    N = 300
    fs = 16000
    fdl = 81
    rir = np.zeros(N)

    time = np.array([100 / fs, 200 / fs])
    alpha = np.array([1.0, 1.0])
    visibility = np.array([1, 1], dtype=np.int32)

    try:
        build_rir.fast_rir_builder(rir, time, alpha[:1], visibility, fs, fdl)
        assert False
    except:
        print("Ok, alpha error occured")
        pass

    try:
        build_rir.fast_rir_builder(rir, time, alpha, visibility[:1], fs, fdl)
        assert False
    except:
        print("Ok, visibility error occured")
        pass

    try:
        build_rir.fast_rir_builder(rir, time, alpha, visibility, fs, 80)
        assert False
    except:
        print("Ok, fdl error occured")
        pass


if __name__ == "__main__":

    import matplotlib.pyplot as plt

    for t, a, v in zip(times, alphas, visibilities):
        ir_ref, ir_cython = build_rir_wrap(
            times[0], alphas[0], visibilities[0], fs, fdl
        )

        print("Error:", np.max(np.abs(ir_ref - ir_cython)))

        plt.figure()
        plt.plot(ir_ref, label="ref")
        plt.plot(ir_cython, label="cython")
        plt.legend()

    test_short()
    test_long()
    test_errors()

    plt.show()



"""
Example of the basic operations with ``pyroomacoustics.datasets.Dataset``
and ``pyroomacoustics.datasets.Sample`` classes
"""
from pyroomacoustics.datasets import Sample, Dataset


def test_dataset():
    # Prepare a few artificial samples
    samples = [
        {
            "data": 0.99,
            "metadata": {
                "speaker": "alice",
                "sex": "female",
                "age": 37,
                "number": "one",
            },
        },
        {
            "data": 2.1,
            "metadata": {
                "speaker": "alice",
                "sex": "female",
                "age": 37,
                "number": "two",
            },
        },
        {
            "data": 1.02,
            "metadata": {"speaker": "bob", "sex": "male", "age": 48, "number": "one"},
        },
        {
            "data": 2.07,
            "metadata": {"speaker": "bob", "sex": "male", "age": 48, "number": "two"},
        },
        {
            "data": 3.0,
            "metadata": {"speaker": "bob", "sex": "male", "age": 48, "number": "three"},
        },
        {
            "data": 1.97,
            "metadata": {
                "speaker": "charles",
                "sex": "male",
                "age": 25,
                "number": "two",
            },
        },
    ]

    dataset = Dataset()
    for s in samples:
        new_sample = Sample(s["data"], **s["metadata"])
        dataset.add_sample(new_sample)

    assert len(dataset) == len(samples)

    assert dataset.info["speaker"]["alice"] == 2
    assert dataset.info["speaker"]["bob"] == 3
    assert dataset.info["speaker"]["charles"] == 1

    assert dataset.info["sex"]["male"] == 4
    assert dataset.info["sex"]["female"] == 2

    assert dataset.info["age"][37] == 2
    assert dataset.info["age"][48] == 3
    assert dataset.info["age"][25] == 1

    assert dataset.info["number"]["one"] == 2
    assert dataset.info["number"]["two"] == 3
    assert dataset.info["number"]["three"] == 1

    assert dataset[2].meta.speaker == "bob"
    assert dataset[2].meta.sex == "male"
    assert dataset[2].meta.age == 48
    assert dataset[2].meta.number == "one"

    # We can obtain a new dataset with only male subject
    dataset2 = dataset.filter(sex="male")

    assert len(dataset2) == 4
    assert dataset2[0].meta.sex == "male"
    assert dataset2[1].meta.sex == "male"
    assert dataset2[2].meta.sex == "male"
    assert dataset2[3].meta.sex == "male"

    # select all samples with age larger than 30
    dataset3 = dataset.filter(age=lambda a: a > 30)
    assert len(dataset3) == 5
    for s in dataset3.samples:
        assert s.meta.age > 30

    # select all samples with a number that contains 't'
    dataset_contains_n = dataset.filter(
        number=["one", "three"], speaker=["alice", "bob"]
    )
    assert len(dataset_contains_n) == 3
    for s in dataset_contains_n.samples:
        assert (s.meta.number == "one" or s.meta.number == "three") and (
            s.meta.speaker == "alice" or s.meta.speaker == "bob"
        )


if __name__ == "__main__":
    test_dataset()



# Test of Wall constructor
# Copyright (C) 2019  Robin, Scheibler, Cyril Cadoux
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
# You should have received a copy of the MIT License along with this program. If
# not, see <https://opensource.org/licenses/MIT>.
from __future__ import division

import numpy as np
import pyroomacoustics as pra

eps = 1e-6

# The vertices of the wall are assumed to turn counter-clockwise around the
# normal of the wall

# A very simple wall
walls = [
    {
        "corners": np.array(
            [
                [0.0, 1.0, 1.0, 0.0],
                [0.0, 0.0, 1.0, 1.0],
                [0.0, 0.0, 0.0, 0.0],
            ]
        ),
        "area": 1,
        "absorption": [0.2],
        "scattering": [0.1],
    },
    {
        "corners": np.array(
            [
                [-1, 1, 1],
                [1, -1, 1],
                [1, 1, -1],
            ]
        ),
        "area": 3.4641016151377557,  # this is an equilateral triangle with side sqrt(8)
        "absorption": [0.2],
        "scattering": [0.1],
    },
]


def test_wall_3d_construct_0():
    """Tests construction of a wall"""
    w_info = walls[0]
    wall = pra.wall_factory(w_info["corners"], [0.2], [0.1])

    return wall


def test_wall_3d_construct_1():
    """Tests construction of a wall"""
    w_info = walls[1]
    wall = pra.wall_factory(w_info["corners"], [0.2], [0.1])

    return wall


def test_wall_3d_area_0():
    """Tests the area computation"""
    w_info = walls[0]
    wall = pra.wall_factory(
        w_info["corners"], w_info["absorption"], w_info["scattering"]
    )
    err = abs(wall.area() - w_info["area"])
    assert err < 1, "The error is {}".format(err)


def test_wall_3d_area_1():
    """Tests the area computation"""
    w_info = walls[1]
    wall = pra.wall_factory(
        w_info["corners"], w_info["absorption"], w_info["scattering"]
    )
    err = abs(wall.area() - w_info["area"])
    assert err < 1, "The error is {}".format(err)


def test_wall_3d_normal_0():
    """Tests direction of normal wrt to point arrangement"""
    w_info = walls[0]
    wall1 = pra.wall_factory(w_info["corners"], [0.2], [0.1])
    # the same wall with normal pointing the other way
    wall2 = pra.wall_factory(w_info["corners"][:, ::-1], [0.2], [0.1])

    err = np.linalg.norm(wall1.normal + wall2.normal)
    assert err < eps, "The error is {}".format(err)


def test_wall_3d_normal_1():
    """Tests direction of normal wrt to point arrangement"""
    w_info = walls[1]
    wall1 = pra.wall_factory(w_info["corners"], [0.2], [0.1])
    # the same wall with normal pointing the other way
    wall2 = pra.wall_factory(w_info["corners"][:, ::-1], [0.2], [0.1])

    err = np.linalg.norm(wall1.normal + wall2.normal)
    assert err < eps, "The error is {}".format(err)


if __name__ == "__main__":

    wall0 = test_wall_3d_construct_0()
    test_wall_3d_normal_0()
    test_wall_3d_area_0()
    wall1 = test_wall_3d_construct_1()
    test_wall_3d_normal_1()
    test_wall_3d_area_1()



# Test of Wall reflection side test
# Copyright (C) 2019  Robin Scheibler, Cyril Cadoux
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
# You should have received a copy of the MIT License along with this program. If
# not, see <https://opensource.org/licenses/MIT>.
from __future__ import division

import unittest
import numpy as np
import pyroomacoustics as pra

eps = 1e-6

corners = {
    "3d": np.array(
        [
            [-1, 1, 1],
            [1, -1, 1],
            [1, 1, -1],
        ]
    ),
    "2d": np.array(
        [
            [-1, 1],
            [-2, 2],
        ]
    ),
}

points = {
    "3d_up": {
        "d": "3d",
        "p": [-1, -1, -1],
        "expect": -1,
        "reflect": [5 / 3, 5 / 3, 5 / 3],
    },
    "3d_down": {
        "d": "3d",
        "p": [1, 1, 1],
        "expect": 1,
        "reflect": [-1 / 3, -1 / 3, -1 / 3],
    },
    "3d_on": {
        "d": "3d",
        "p": [1 / 3, 1 / 3, 1 / 3],
        "expect": 0,
        "reflect": [1 / 3, 1 / 3, 1 / 3],
    },
    "2d_down": {
        "d": "2d",
        "p": [-2, 1],
        "expect": -1,
        "reflect": [2, -1],
    },
    "2d_up": {
        "d": "2d",
        "p": [2, -1],
        "expect": 1,
        "reflect": [-2, 1],
    },
    "2d_on": {
        "d": "2d",
        "p": [0.5, 1],
        "expect": 0,
        "reflect": [0.5, 1],
    },
}


def run_side(label):
    p = points[label]["p"]
    r_exp = points[label]["expect"]
    d = points[label]["d"]

    wall = pra.wall_factory(corners[d], [0.1], [0.1])
    r = wall.side(p)

    print("{}: returned={} expected={}".format(label, r, r_exp))
    return r == r_exp


def run_reflect(label):

    p = points[label]["p"]
    p_refl = np.array(points[label]["reflect"])
    r_exp = points[label]["expect"]
    d = points[label]["d"]

    wall = pra.wall_factory(corners[d], [0.1], [0.1])

    x = np.zeros(wall.dim, dtype=np.float32)
    wall.reflect(p, x)

    err = np.linalg.norm(x - p_refl) < eps

    print("{}: error={}".format(label, err))
    return err


class TestUtilityRoutines(unittest.TestCase):
    def test_side_3d_up(self):
        ret = run_side("3d_up")
        self.assertTrue(ret)

    def test_side_3d_down(self):
        ret = run_side("3d_down")
        self.assertTrue(ret)

    def test_side_3d_on(self):
        ret = run_side("3d_on")
        self.assertTrue(ret)

    def test_side_2d_up(self):
        ret = run_side("2d_up")
        self.assertTrue(ret)

    def test_side_2d_down(self):
        ret = run_side("2d_down")
        self.assertTrue(ret)

    def test_side_2d_on(self):
        ret = run_side("2d_on")
        self.assertTrue(ret)

    def test_reflect_3d_up(self):
        ret = run_reflect("3d_up")
        self.assertTrue(ret)

    def test_reflect_3d_down(self):
        ret = run_reflect("3d_down")
        self.assertTrue(ret)

    def test_reflect_3d_on(self):
        ret = run_reflect("3d_on")
        self.assertTrue(ret)

    def test_reflect_2d_up(self):
        ret = run_reflect("2d_up")
        self.assertTrue(ret)

    def test_reflect_2d_down(self):
        ret = run_reflect("2d_down")
        self.assertTrue(ret)

    def test_reflect_2d_on(self):
        ret = run_reflect("2d_on")
        self.assertTrue(ret)

    def test_reflected_end2D(self):

        eps = 0.001
        start = [1, 3]

        hit = [5, 3]
        # normal = [-1, -1]
        corners = np.array(
            [
                [6, 4],
                [4, 6],
            ]
        )
        wall = pra.wall_factory(corners, [0.1], [0.1])
        length = 4

        res = wall.normal_reflect(start, hit, length)
        self.assertTrue(np.allclose(res, [5.0, -1.0], atol=eps))

    def test_reflected_end3D(self):

        eps = 0.001
        start = [1, 1, 1]
        hit = [-1, 1, 3]
        # normal = [1, 0, 0]
        corners = np.array(
            [
                [
                    -1,
                    -1,
                    -1,
                    -1,
                ],
                [
                    0,
                    2,
                    2,
                    0,
                ],
                [
                    2,
                    2,
                    4,
                    4,
                ],
            ]
        )
        wall = pra.wall_factory(corners, [0.1], [0.1])

        length = 2 * np.sqrt(2)

        res = wall.normal_reflect(start, hit, length)
        self.assertTrue(np.allclose(res, [1, 1, 5], atol=eps))


if __name__ == "__main__":
    unittest.main()



# Test of polygon point inclusion test
# Copyright (C) 2019  Robin Scheibler, Cyril Cadoux, Sidney Barthe
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
# You should have received a copy of the MIT License along with this program. If
# not, see <https://opensource.org/licenses/MIT>.
from __future__ import division

import numpy as np
import pyroomacoustics as pra

polygons = [
    np.array(
        [  # this one is clockwise
            [
                0,
                4,
                4,
                0,
            ],
            [
                0,
                0,
                4,
                4,
            ],
        ]
    ),
    np.array(
        [  # this one is clockwise!
            [0, 0, 1, 1, 3, 3],
            [0, 1, 1, 2, 2, 0],
        ]
    ),
    np.array(
        [  # this one is clockwise!
            [0, 1, 1, 3, 3],
            [0, 1, 2, 2, 0],
        ]
    ),
]

cases = {
    "inside": {
        "pol": 0,
        "p": [2, 2],
        "ret": 0,
    },
    "on_border": {
        "pol": 0,
        "p": [0, 2],
        "ret": 1,
    },
    "on_corner": {
        "pol": 0,
        "p": [4, 4],
        "ret": 1,
    },
    "outside": {
        "pol": 0,
        "p": [5, 5],
        "ret": -1,
    },
    # horizontal wall aligned with point
    "horiz_wall_align": {
        "pol": 1,
        "p": [2, 1],
        "ret": 0,
    },
    # ray is going through vertex
    "ray_through_vertex": {
        "pol": 2,
        "p": [2, 1],
        "ret": 0,
    },
    # point is at the same height as top of polygon, but outside
    "top_outside": {
        "pol": 2,
        "p": [4, 2],
        "ret": -1,
    },
}


def run_inside_pol(lbl):
    pol = polygons[cases[lbl]["pol"]]
    p = cases[lbl]["p"]
    r_exp = cases[lbl]["ret"]

    ret = pra.libroom.is_inside_2d_polygon(p, pol)

    assert ret == r_exp, "{} : returned={} expected={}".format(lbl, ret, r_exp)


def test_inside():
    run_inside_pol("inside")


def test_on_border():
    run_inside_pol("on_border")


def test_on_corner():
    run_inside_pol("on_corner")


def test_outside():
    run_inside_pol("outside")


def test_horiz_wall_align():
    run_inside_pol("horiz_wall_align")


def test_ray_through_vertex():
    run_inside_pol("ray_through_vertex")


def test_top_outside():
    run_inside_pol("top_outside")


if __name__ == "__main__":

    test_inside()
    test_on_border()
    test_on_corner()
    test_outside()
    test_top_outside()
    test_horiz_wall_align()
    test_ray_through_vertex()



# @version: 1.0  date: 05/06/2015 by Sidney Barthe
# @author: robin.scheibler@epfl.ch, ivan.dokmanic@epfl.ch, sidney.barthe@epfl.ch
# @copyright: EPFL-IC-LCAV 2015

from unittest import TestCase

import numpy as np
import pyroomacoustics as pra

fs = 8000
t0 = 0.0
absorption = 0.90
e_abs = 1.0 - (1.0 - absorption) ** 2
max_order_sim = 2
sigma2_n = 5e-7

corners = np.array(
    [
        [0, 5, 5, 3, 3, 2, 2, 0],
        [0, 0, 5, 5, 2, 2, 5, 5],
    ]
)
room = pra.Room.from_corners(
    corners,
    materials=pra.Material(e_abs),
    fs=fs,
    t0=t0,
    max_order=max_order_sim,
    sigma2_awgn=sigma2_n,
)

room.add_source([1, 4], None, 0)

# place 3 microphones in the room
mics = pra.MicrophoneArray(
    np.array(
        [
            [
                4,
                2.5,
                2.5,
            ],
            [4, 1, 4],
        ]
    ),
    fs,
)
room.add_microphone_array(mics)

# run the image source model
room.image_source_model()

# we order the sources lexicographically
ordering = np.lexsort(room.sources[0].images)
images_found = room.sources[0].images[:, ordering]
visibilities = room.visibility[0][:, ordering]

# This should be all visible sources, from the 3 mic locations
sources = np.array(
    [[1.0, -1.0, 1.0, -3.0, -1.0, -1.0, 1.0], [-6.0, -4.0, -4.0, 4.0, 4.0, 6.0, 8.0]]
)

# These are the visibilities for each individual microphone
visible_opposite_side = [0, 0, 1, 0, 0, 0, 0]
visible_middle = [1, 1, 1, 1, 1, 1, 1]
visible_outside = [0, 0, 0, 0, 0, 0, 0]


def test_sources():
    assert np.allclose(images_found, sources)


def test_visibility_oppositeSide():
    assert all(visibilities[0] == visible_opposite_side)


def test_visibility_middle():
    assert all(visibilities[1] == visible_middle)


def test_visibility_outside():
    assert all(visibilities[2] == visible_outside)


if __name__ == "__main__":
    test_sources()
    test_visibility_oppositeSide()
    test_visibility_middle()
    test_visibility_outside()



"""
Created on Tue Feb  8 17:50:45 2022

@author: od0014

Script to test removal of sweeping echoes with randomized image method
"""
import pyroomacoustics as pra
from pyroomacoustics import metrics as met


# create an example with sweeping echo - from Enzo's paper
room_size = [4, 4, 4]
source_loc = [1, 2, 2]
mic_loc = [0.5, 1, 0.75]
fs = 44100


def test_ism():

    room = pra.ShoeBox(room_size, fs, materials=pra.Material(0.1), max_order=50)

    room.add_source(source_loc)

    room.add_microphone(mic_loc)

    room.compute_rir()

    ssf_ism = met.sweeping_echo_measure(room.rir[0][0], fs)

    assert 0 <= ssf_ism <= 1.0

    return ssf_ism


def test_random_ism():

    room = pra.ShoeBox(
        room_size, fs, materials=pra.Material(0.1), max_order=50, use_rand_ism=True
    )

    room.add_source(source_loc)

    room.add_microphone(mic_loc)

    room.compute_rir()

    # measure of spectral flatness of sweeping echoes
    # higher value is desired
    ssf_rism = met.sweeping_echo_measure(room.rir[0][0], fs)

    assert 0 <= ssf_rism <= 1.0

    return ssf_rism


if __name__ == "__main__":
    ssf_ism = test_ism()
    ssf_rism = test_random_ism()
    assert ssf_rism > ssf_ism



from .music import *


class NormMUSIC(MUSIC):
    """
    Class to apply MUltiple SIgnal Classication (MUSIC) direction-of-arrival
    (DoA) with frequency normalization for a particular microphone array.

    .. note:: Run locate_source() to apply the NormMUSIC algorithm.

    Parameters
    ----------
    L: numpy array
        Microphone array positions. Each column should correspond to the
        cartesian coordinates of a single microphone.
    fs: float
        Sampling frequency.
    nfft: int
        FFT length.
    c: float
        Speed of sound. Default: 343 m/s
    num_src: int
        Number of sources to detect. Default: 1
    mode: str
        'far' or 'near' for far-field or near-field detection
        respectively. Default: 'far'
    r: numpy array
        Candidate distances from the origin. Default: np.ones(1)
    azimuth: numpy array
        Candidate azimuth angles (in radians) with respect to x-axis.
        Default: np.linspace(-180.,180.,30)*np.pi/180
    colatitude: numpy array
        Candidate elevation angles (in radians) with respect to z-axis.
        Default is x-y plane search: np.pi/2*np.ones(1)
    frequency_normalization: bool
        If True, the MUSIC pseudo-spectra are normalized before averaging across the frequency axis, default:True
    """

    def __init__(
        self,
        L,
        fs,
        nfft,
        c=343.0,
        num_src=1,
        mode="far",
        r=None,
        azimuth=None,
        colatitude=None,
        frequency_normalization=True,
        **kwargs
    ):
        super().__init__(
            L,
            fs,
            nfft,
            c,
            num_src,
            mode,
            r,
            azimuth,
            colatitude,
            frequency_normalization,
            **kwargs
        )



from __future__ import division, print_function

import numpy as np
import scipy.linalg as la
from scipy import fftpack

try:
    import mkl_fft as fft

    has_mklfft = True
except ImportError:
    import numpy.fft as fft

    has_mklfft = False


def autocorr(x):
    """Fast autocorrelation computation using the FFT"""

    X = fft.rfft(x, n=2 * x.shape[0])
    r = fft.irfft(np.abs(X) ** 2)

    return r[: x.shape[0]] / (x.shape[0] - 1)


def toeplitz_multiplication(c, r, A, **kwargs):
    """
    Compute numpy.dot(scipy.linalg.toeplitz(c,r), A) using the FFT.

    Parameters
    ----------
    c: ndarray
        the first column of the Toeplitz matrix
    r: ndarray
        the first row of the Toeplitz matrix
    A: ndarray
        the matrix to multiply on the right
    """

    m = c.shape[0]
    n = r.shape[0]

    fft_len = int(2 ** np.ceil(np.log2(m + n - 1)))
    zp = fft_len - m - n + 1

    if A.shape[0] != n:
        raise ValueError("A dimensions not compatible with toeplitz(c,r)")

    x = np.concatenate((c, np.zeros(zp, dtype=c.dtype), r[-1:0:-1]))
    xf = np.fft.rfft(x, n=fft_len)

    Af = np.fft.rfft(A, n=fft_len, axis=0)

    return np.fft.irfft((Af.T * xf).T, n=fft_len, axis=0)[
        :m,
    ]


def hankel_multiplication(c, r, A, mkl=True, **kwargs):
    """
    Compute numpy.dot(scipy.linalg.hankel(c,r=r), A) using the FFT.

    Parameters
    ----------
    c: ndarray
        the first column of the Hankel matrix
    r: ndarray
        the last row of the Hankel matrix
    A: ndarray
        the matrix to multiply on the right
    mkl: bool, optional
        if True, use the mkl_fft package if available
    """

    if mkl and has_mklfft:
        fmul = mkl_toeplitz_multiplication
    else:
        fmul = toeplitz_multiplication
        A = A[: r.shape[0], :]

    return fmul(c[::-1], r, A, **kwargs)[
        ::-1,
    ]


def mkl_toeplitz_multiplication(c, r, A, A_padded=False, out=None, fft_len=None):
    """
    Compute numpy.dot(scipy.linalg.toeplitz(c,r), A) using the FFT from the mkl_fft package.

    Parameters
    ----------
    c: ndarray
        the first column of the Toeplitz matrix
    r: ndarray
        the first row of the Toeplitz matrix
    A: ndarray
        the matrix to multiply on the right
    A_padded: bool, optional
        the A matrix can be pre-padded with zeros by the user, if this is the case
        set to True
    out: ndarray, optional
        an ndarray to store the output of the multiplication
    fft_len: int, optional
        specify the length of the FFT to use
    """

    if not has_mklfft:
        raise ValueError(
            "Import mkl_fft package unavailable. Install from https://github.com/LCAV/mkl_fft"
        )

    m = c.shape[0]
    n = r.shape[0]

    if fft_len is None:
        fft_len = int(2 ** np.ceil(np.log2(m + n - 1)))
    zp = fft_len - m - n + 1

    if (not A_padded and A.shape[0] != n) or (A_padded and A.shape[0] != fft_len):
        raise ValueError("A dimensions not compatible with toeplitz(c,r)")

    x = np.concatenate((c, np.zeros(zp, dtype=c.dtype), r[-1:0:-1]))
    xf = fft.rfft(x, n=fft_len)

    if out is not None:
        fft.rfft(A, n=fft_len, axis=0, out=out)
    else:
        out = fft.rfft(A, n=fft_len, axis=0)

    out *= xf[:, None]

    if A_padded:
        fft.irfft(out, n=fft_len, axis=0, out=A)
    else:
        A = fft.irfft(out, n=fft_len, axis=0)

    return A[
        :m,
    ]


def naive_toeplitz_multiplication(c, r, A):
    """
    Compute numpy.dot(scipy.linalg.toeplitz(c,r), A)

    Parameters
    ----------
    c: ndarray
        the first column of the Toeplitz matrix
    r: ndarray
        the first row of the Toeplitz matrix
    A: ndarray
        the matrix to multiply on the right
    """

    return np.dot(la.toeplitz(c, r), A)


def hankel_stride_trick(x, shape):
    """
    Make a Hankel matrix from a vector using stride tricks

    Parameters
    ----------
    x: ndarray
        a vector that contains the concatenation of the first column
        and first row of the Hankel matrix to build *without* repetition
        of the lower left corner value of the matrix
    shape: tuple
        the shape of the Hankel matrix to build, it must satisfy ``x.shape[0] == shape[0] + shape[1] - 1``
    """

    if x.shape[0] != shape[0] + shape[1] - 1:
        raise ValueError("Inconsistent dimensions")

    strides = (x.itemsize, x.itemsize)
    return np.lib.stride_tricks.as_strided(x, shape=shape, strides=strides)


def toeplitz_strang_circ_approx(r, matrix=False):
    """
    Circulant approximation to a symetric Toeplitz matrix
    by Gil Strang

    Parameters
    ----------
    r: ndarray
        the first row of the symmetric Toeplitz matrix
    matrix: bool, optional
        if True, the full symetric Toeplitz matrix is returned,
        otherwise, only the first column
    """

    n = r.shape[0]
    c = r.copy()
    m = n // 2 if n % 2 == 0 else (n - 1) // 2
    c[-m:] = r[m:0:-1]

    if matrix:
        return la.circulant(c)
    else:
        return c


def toeplitz_opt_circ_approx(r, matrix=False):
    """
    Optimal circulant approximation of a symmetric Toeplitz matrix
    by Tony F. Chan

    Parameters
    ----------
    r: ndarray
        the first row of the symmetric Toeplitz matrix
    matrix: bool, optional
        if True, the full symetric Toeplitz matrix is returned,
        otherwise, only the first column
    """

    n = r.shape[0]

    r_rev = np.zeros(r.shape)
    r_rev[1:] = r[:0:-1]

    i = np.arange(n)
    c = (i * r_rev + (n - i) * r) / n
    c[1:] = c[:0:-1]

    if matrix:
        return la.circulant(c)
    else:
        return c


if __name__ == "__main__":

    import time

    try:
        import mkl as mkl_service

        mkl_service.set_num_threads(7)
    except ImportError:
        pass

    n_iter = 1

    m = 10
    n = 100
    o = 10

    dtype_in = np.float32
    dtype_out = np.complex128 if dtype_in == np.float64 else np.complex64
    dorder = "C"
    fft_len = int(2 ** np.ceil(np.log2(m + n - 1)))

    c = dtype_in(np.random.randn(m))
    r = dtype_in(np.random.randn(n))

    A = np.asfortranarray(dtype_in(np.random.randn(n, o)))

    Apad = np.zeros((fft_len, o), order=dorder, dtype=dtype_in)
    Apad[:n, :] = A
    out = np.zeros((fft_len // 2 + 1, o), dtype=dtype_out, order=dorder)

    print("Start...")
    start_time = time.time()
    for i in range(n_iter):
        C1 = naive_toeplitz_multiplication(c, r, A)
    nai_time = time.time() - start_time
    print("naive --- %s seconds ---" % (nai_time))

    """
    start_time = time.time()
    for i in range(n_iter):
        C = toeplitz_multiplication(c, r, A)
    npy_time = time.time() - start_time
    print("numpy   --- %s seconds ---" % (npy_time))
    """

    start_time = time.time()
    for i in range(n_iter):
        # C2 = mkl_toeplitz_multiplication(c, r, Apad, A_padded=True, out=out, fft_len=fft_len)
        C2 = mkl_toeplitz_multiplication(
            c, r, A, A_padded=False, out=None, fft_len=None
        )
    mkl_time = time.time() - start_time

    print("mkl     --- %s seconds ---" % (mkl_time))
    print("Result matching:", np.allclose(C1, C2))

    print("Stop.")



import numpy as np
import pyroomacoustics as pra


def test_anechoic_room(debug=False):

    # sound signal
    fs = 40000
    freq = 440
    times = np.linspace(0, 1, 100)
    signal = np.sin(2 * np.pi * freq * times)

    for dim in [2, 3]:

        # create Anechoic room using the correct class
        room_infinite = pra.AnechoicRoom(fs=fs, dim=dim)

        # create same room using Shoebox
        room_shoebox = pra.ShoeBox([10] * dim, fs=fs, max_order=0)

        # Add a source somewhere in the room
        soundsource = pra.SoundSource([5] * dim, signal=signal)
        room_infinite.add_soundsource(soundsource)
        room_shoebox.add_soundsource(soundsource)

        # Create a linear array beamformer with 4 microphones
        # with angle 0 degrees and inter mic distance 10 cm
        R = pra.linear_2D_array([2, 1.5], 4, 0, 0.04)
        if dim == 3:
            R = np.r_[R, np.zeros((1, R.shape[1]))]

        beamformer = pra.Beamformer(R, fs)

        room_infinite.add_microphone_array(beamformer)
        room_shoebox.add_microphone_array(beamformer)

        # Now compute the delay and sum weights for the beamformer
        for room in [room_infinite, room_shoebox]:
            room.mic_array.rake_delay_and_sum_weights(room.sources[0][:1])
            if debug:
                room.plot(freq=[1000, 2000, 4000, 8000], img_order=0)

        # create signal
        room_infinite.sources[0].add_signal(signal)

        # make sure that the premix signals for both are the same.
        premix_infinite = room_infinite.simulate(return_premix=True)
        premix_shoebox = room_shoebox.simulate(return_premix=True)

        if debug:
            plt.figure()
            for m in range(premix_infinite.shape[1]):
                plt.plot(
                    premix_infinite[0][m],
                    label="mic {} infinite".format(m),
                    color="C{}".format(m),
                )
                plt.plot(
                    premix_shoebox[0][m],
                    label="mic {} shoebox".format(m),
                    ls=":",
                    color="C{}".format(m),
                )
            plt.legend()
        else:
            np.testing.assert_allclose(premix_infinite, premix_shoebox)
    if debug:
        plt.show()


if __name__ == "__main__":
    import matplotlib.pylab as plt

    test_anechoic_room(debug=True)



"""
This tests the construction of a bank of octave filters
"""
import pyroomacoustics as pra

import numpy as np
from scipy.signal import sosfreqz

tol = 1.0  # decibel


def test_bandpass_filterbank():
    f0 = 125.0 / 2.0
    fs = 16000

    bands, fc = pra.octave_bands(f0)
    filters = pra.bandpass_filterbank(bands, fs=fs, order=16)

    res = None
    for sos, f in zip(filters, fc):
        w, h = sosfreqz(sos, worN=4096, whole=False)

        # inspect the cumulative response
        if res is None:
            res = np.abs(h) ** 2
        else:
            res += np.abs(h) ** 2

    # check automatically that the cumulative response is (nearly) flat
    freq = w / np.pi * fs / 2
    I = np.where(
        np.logical_and(freq > fc[0], freq < np.minimum(fc[-1], 0.95 * fs / 2))
    )[0]
    err = np.max(np.abs(10 * np.log10(res[I])))

    assert err <= tol, "Bandpass filterbank test fails (err {} > tol {})".format(
        err, tol
    )


if __name__ == "__main__":

    import matplotlib.pyplot as plt

    f0 = 125.0 / 2.0
    fs = 16000

    bands, fc = pra.octave_bands(f0)
    filters = pra.bandpass_filterbank(bands, fs=fs, order=16)

    plt.figure()
    res = None
    for sos, f in zip(filters, fc):
        w, h = sosfreqz(sos, worN=4096, whole=False)
        plt.semilogx(
            w / np.pi * fs / 2.0, 20 * np.log10(np.abs(h)), label=str(f) + " Hz"
        )

        # inspect the cumulative response
        if res is None:
            res = np.abs(h) ** 2
        else:
            res += np.abs(h) ** 2

    # check automatically that the cumulative response is (nearly) flat
    freq = w / np.pi * fs / 2
    I = np.where(
        np.logical_and(freq > fc[0], freq < np.minimum(fc[-1], 0.95 * fs / 2))
    )[0]
    err = np.max(np.abs(10 * np.log10(res[I])))
    print("The error is", err, "(tol is", tol, "dB)")

    plt.semilogx(w / np.pi * fs / 2.0, 10 * np.log10(np.abs(res)), label="Sum")

    plt.title("Octave Filter Bank")
    plt.xlabel("Frequency [Hz]")
    plt.ylabel("Amplitude [dB]")
    plt.margins(0, 0.1)
    plt.grid(which="both", axis="both")
    plt.axvline(100, color="green")  # cutoff frequency
    plt.ylim(-60, 3)
    plt.xlim(10.0, fs / 2.0)
    plt.legend()
    plt.show()

    # run the automatic tests
    test_bandpass_filterbank()



from __future__ import division, print_function
import numpy as np
import pyroomacoustics as pra
import time

try:
    import pyfftw

    pyfftw_available = True
except ImportError:
    pyfftw_available = False

try:
    import mkl_fft

    mkl_available = True
except ImportError:
    mkl_available = False


n_trials = 1000
nfft = 128
D = 7
x = np.random.randn(nfft, D).astype("float32")


def timing(transform, n_trials):

    dft = pra.transform.DFT(nfft, D, transform=transform)
    start_time = time.time()
    for k in range(n_trials):
        dft.analysis(x)
    analysis_time = (time.time() - start_time) / n_trials * 1e6
    start_time = time.time()
    for k in range(n_trials):
        dft.synthesis()
    synthesis_time = (time.time() - start_time) / n_trials * 1e6

    print(
        "avg %s : %f [1e-6 sec], (analysis, synthesis)=(%f, %f) [1e-6 sec]"
        % (transform, analysis_time + synthesis_time, analysis_time, synthesis_time)
    )


res = timing("numpy", n_trials)
if pyfftw_available:
    res = timing("fftw", n_trials)
if mkl_available:
    res = timing("mkl", n_trials)

"""
test against without using class
"""
print()
start_time = time.time()
for k in range(n_trials):
    X = np.fft.rfft(x)
analysis_time = (time.time() - start_time) / n_trials * 1e6
start_time = time.time()
for k in range(n_trials):
    x_r = np.fft.irfft(X)
synthesis_time = (time.time() - start_time) / n_trials * 1e6
print(
    "avg numpy w/o class : %f [1e-6 sec], (analysis, synthesis)=(%f, %f) [1e-6 sec]"
    % (analysis_time + synthesis_time, analysis_time, synthesis_time)
)

if pyfftw_available:

    # prepare
    a = pyfftw.empty_aligned([nfft, D], dtype="float32")
    b = pyfftw.empty_aligned([nfft // 2 + 1, D], dtype="complex64")
    c = pyfftw.empty_aligned([nfft, D], dtype="float32")
    forward = pyfftw.FFTW(a, b, axes=(0,))
    backward = pyfftw.FFTW(b, c, axes=(0,), direction="FFTW_BACKWARD")

    start_time = time.time()
    for k in range(n_trials):
        forward()
    analysis_time = (time.time() - start_time) / n_trials * 1e6
    start_time = time.time()
    for k in range(n_trials):
        backward()
    synthesis_time = (time.time() - start_time) / n_trials * 1e6
    print(
        "avg fftw w/o class : %f [1e-6 sec], (analysis, synthesis)=(%f, %f) [1e-6 sec]"
        % (analysis_time + synthesis_time, analysis_time, synthesis_time)
    )

if mkl_available:
    start_time = time.time()
    for k in range(n_trials):
        X = mkl_fft.rfft_numpy(x)
    analysis_time = (time.time() - start_time) / n_trials * 1e6
    start_time = time.time()
    for k in range(n_trials):
        x_r = mkl_fft.irfft_numpy(X)
    synthesis_time = (time.time() - start_time) / n_trials * 1e6
    print(
        "avg mkl w/o class : %f [1e-6 sec], (analysis, synthesis)=(%f, %f) [1e-6 sec]"
        % (analysis_time + synthesis_time, analysis_time, synthesis_time)
    )



# @version: 1.0  date: 05/06/2015 by Sidney Barthe
# @author: robin.scheibler@epfl.ch, ivan.dokmanic@epfl.ch, sidney.barthe@epfl.ch
# @copyright: EPFL-IC-LCAV 2015

import unittest

import numpy as np
import pyroomacoustics as pra

fs = 8000
t0 = 0.0
absorption = 0.90
e_abs = 1.0 - (1.0 - absorption) ** 2
max_order_sim = 2
sigma2_n = 5e-7

corners = np.array([[0, 0, 6, 6], [0, 6, 6, 0]])
room = pra.Room.from_corners(
    corners,
    materials=pra.Material(e_abs),
    fs=fs,
    t0=t0,
    max_order=max_order_sim,
    sigma2_awgn=sigma2_n,
)

room.extrude(6)

room.add_source([3, 3, 3], None, 0)

# place 3 microphones in the room
mics = pra.MicrophoneArray(np.array([[3, 5, 0, 7], [3, 3, 0, 7], [3, 3, 0, 7]]), fs)
room.add_microphone_array(mics)

# run the image source model
room.image_source_model()

# we order the sources lexicographically
ordering = np.lexsort(room.sources[0].images)
images_found = room.sources[0].images[:, ordering]
visibilities = np.array(room.visibility[0][:, ordering], dtype=bool)

sources = np.array(
    [
        [
            3.0,
            3.0,
            3.0,
            -3.0,
            -3.0,
            3.0,
            9.0,
            9.0,
            3.0,
            3.0,
            3.0,
            -3.0,
            -3.0,
            3.0,
            9.0,
            9.0,
            -9.0,
            -3.0,
            3.0,
            9.0,
            15.0,
            -3.0,
            -3.0,
            3.0,
            9.0,
            9.0,
            3.0,
            3.0,
            3.0,
            -3.0,
            -3.0,
            3.0,
            9.0,
            9.0,
            3.0,
            3.0,
            3.0,
        ],
        [
            3.0,
            -3.0,
            -3.0,
            3.0,
            3.0,
            3.0,
            3.0,
            3.0,
            9.0,
            9.0,
            -9.0,
            -3.0,
            -3.0,
            -3.0,
            -3.0,
            -3.0,
            3.0,
            3.0,
            3.0,
            3.0,
            3.0,
            9.0,
            9.0,
            9.0,
            9.0,
            9.0,
            15.0,
            -3.0,
            -3.0,
            3.0,
            3.0,
            3.0,
            3.0,
            3.0,
            9.0,
            9.0,
            3.0,
        ],
        [
            -9.0,
            -3.0,
            -3.0,
            -3.0,
            -3.0,
            -3.0,
            -3.0,
            -3.0,
            -3.0,
            -3.0,
            3.0,
            3.0,
            3.0,
            3.0,
            3.0,
            3.0,
            3.0,
            3.0,
            3.0,
            3.0,
            3.0,
            3.0,
            3.0,
            3.0,
            3.0,
            3.0,
            3.0,
            9.0,
            9.0,
            9.0,
            9.0,
            9.0,
            9.0,
            9.0,
            9.0,
            9.0,
            15.0,
        ],
    ]
)

visible_middle = np.array(
    [
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
    ],
    dtype=bool,
)
visible_east = np.array(
    [
        1,
        1,
        1,
        1,
        0,
        1,
        0,
        1,
        1,
        1,
        1,
        1,
        0,
        1,
        0,
        1,
        1,
        1,
        1,
        1,
        1,
        0,
        1,
        1,
        0,
        1,
        1,
        1,
        1,
        1,
        0,
        1,
        0,
        1,
        1,
        1,
        1,
    ],
    dtype=bool,
)
visible_bottomCorner = np.array(
    [
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        0,
        1,
        0,
        1,
        1,
        1,
        1,
        1,
        0,
        1,
        1,
        1,
        1,
        1,
        1,
        0,
        1,
        1,
        1,
        1,
        0,
        1,
        0,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
    ],
    dtype=bool,
)
visible_outside = np.array(
    [
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
    ],
    dtype=bool,
)


class TestVisibilityShoeBox3D(unittest.TestCase):
    def test_sources(self):
        self.assertTrue(np.allclose(images_found, sources))

    def test_visibility_middle(self):
        self.assertTrue(
            np.allclose(images_found[:, visibilities[0, :]], sources[:, visible_middle])
        )

    def test_visibility_east(self):
        self.assertTrue(
            np.allclose(images_found[:, visibilities[1, :]], sources[:, visible_east])
        )

    def test_visibility_southWestBottomCorner(self):
        self.assertTrue(
            np.allclose(
                images_found[:, visibilities[2, :]], sources[:, visible_bottomCorner]
            )
        )

    def test_visibility_outside(self):
        self.assertTrue(
            np.allclose(
                images_found[:, visibilities[3, :]], sources[:, visible_outside]
            )
        )


if __name__ == "__main__":
    unittest.main()



"""
A collection of functions to plot maps and points on circles and spheres.
"""
import numpy as np


def polar_plt_dirac(
    self,
    azimuth_ref=None,
    alpha_ref=None,
    save_fig=False,
    file_name=None,
    plt_dirty_img=True,
):
    """
    Generate polar plot of DoA results.

    Parameters
    ----------
    azimuth_ref: numpy array
        True direction of sources (in radians).
    alpha_ref: numpy array
        Estimated amplitude of sources.
    save_fig: bool
        Whether or not to save figure as pdf.
    file_name: str
        Name of file (if saved). Default is
        'polar_recon_dirac.pdf'
    plt_dirty_img: bool
        Whether or not to plot spatial spectrum or
        'dirty image' in the case of FRI.
    """

    try:
        import matplotlib.pyplot as plt
    except ImportError:
        import warnings

        warnings.warn("Matplotlib is required for plotting")
        return

    if self.dim != 2:
        raise ValueError("This function only handles 2D problems.")

    azimuth_recon = self.azimuth_recon
    num_mic = self.M
    phi_plt = self.grid.azimuth

    # determine amplitudes
    from fri import FRI

    if not isinstance(self, FRI):  # use spatial spectrum

        dirty_img = self.grid.values
        alpha_recon = self.grid.values[self.src_idx]
        alpha_ref = alpha_recon

    else:  # create dirty image

        dirty_img = self._gen_dirty_img()
        alpha_recon = np.mean(np.abs(self.alpha_recon), axis=1)
        alpha_recon /= alpha_recon.max()
        if alpha_ref is None:  # non-simulated case
            alpha_ref = alpha_recon

    # plot
    fig = plt.figure(figsize=(5, 4), dpi=90)
    ax = fig.add_subplot(111, projection="polar")
    base = 1.0
    height = 10.0
    blue = [0, 0.447, 0.741]
    red = [0.850, 0.325, 0.098]

    if azimuth_ref is not None:
        if alpha_ref.shape[0] < azimuth_ref.shape[0]:
            alpha_ref = np.concatenate(
                (alpha_ref, np.zeros(azimuth_ref.shape[0] - alpha_ref.shape[0]))
            )

        # match detected with truth
        recon_err, sort_idx = polar_distance(azimuth_recon, azimuth_ref)
        if self.num_src > 1:
            azimuth_recon = azimuth_recon[sort_idx[:, 0]]
            alpha_recon = alpha_recon[sort_idx[:, 0]]
            azimuth_ref = azimuth_ref[sort_idx[:, 1]]
            alpha_ref = alpha_ref[sort_idx[:, 1]]
        elif azimuth_ref.shape[0] > 1:  # one detected source
            alpha_ref[sort_idx[1]] = alpha_recon

        # markers for original doa
        K = len(azimuth_ref)
        ax.scatter(
            azimuth_ref,
            base + height * alpha_ref,
            c=np.tile(blue, (K, 1)),
            s=70,
            alpha=0.75,
            marker="^",
            linewidths=0,
            label="original",
        )

        # stem for original doa
        if K > 1:
            for k in range(K):
                ax.plot(
                    [azimuth_ref[k], azimuth_ref[k]],
                    [base, base + height * alpha_ref[k]],
                    linewidth=1.5,
                    linestyle="-",
                    color=blue,
                    alpha=0.6,
                )
        else:
            ax.plot(
                [azimuth_ref, azimuth_ref],
                [base, base + height * alpha_ref],
                linewidth=1.5,
                linestyle="-",
                color=blue,
                alpha=0.6,
            )

    K_est = azimuth_recon.size

    # markers for reconstructed doa
    ax.scatter(
        azimuth_recon,
        base + height * alpha_recon,
        c=np.tile(red, (K_est, 1)),
        s=100,
        alpha=0.75,
        marker="*",
        linewidths=0,
        label="reconstruction",
    )

    # stem for reconstructed doa
    if K_est > 1:
        for k in range(K_est):
            ax.plot(
                [azimuth_recon[k], azimuth_recon[k]],
                [base, base + height * alpha_recon[k]],
                linewidth=1.5,
                linestyle="-",
                color=red,
                alpha=0.6,
            )

    else:
        ax.plot(
            [azimuth_recon, azimuth_recon],
            [base, base + height * alpha_recon],
            linewidth=1.5,
            linestyle="-",
            color=red,
            alpha=0.6,
        )

    # plot the 'dirty' image
    if plt_dirty_img:
        dirty_img = np.abs(dirty_img)
        min_val = dirty_img.min()
        max_val = dirty_img.max()
        dirty_img = (dirty_img - min_val) / (max_val - min_val)

        # we need to make a complete loop, copy first value to last
        c_phi_plt = np.r_[phi_plt, phi_plt[0]]
        c_dirty_img = np.r_[dirty_img, dirty_img[0]]
        ax.plot(
            c_phi_plt,
            base + height * c_dirty_img,
            linewidth=1,
            alpha=0.55,
            linestyle="-",
            color=[0.466, 0.674, 0.188],
            label="spatial spectrum",
        )

    handles, labels = ax.get_legend_handles_labels()
    ax.legend(
        handles=handles[:3],
        framealpha=0.5,
        scatterpoints=1,
        loc=8,
        fontsize=9,
        ncol=1,
        bbox_to_anchor=(0.9, -0.17),
        handletextpad=0.2,
        columnspacing=1.7,
        labelspacing=0.1,
    )

    ax.set_xlabel(r"azimuth $\bm{\varphi}$", fontsize=11)
    ax.set_xticks(np.linspace(0, 2 * np.pi, num=12, endpoint=False))
    ax.xaxis.set_label_coords(0.5, -0.11)
    ax.set_yticks(np.linspace(0, 1, 2))
    ax.xaxis.grid(b=True, color=[0.3, 0.3, 0.3], linestyle=":")
    ax.yaxis.grid(b=True, color=[0.3, 0.3, 0.3], linestyle="--")
    ax.set_ylim([0, 1.05 * (base + height)])
    if save_fig:
        if file_name is None:
            file_name = "polar_recon_dirac.pdf"
        plt.savefig(file_name, format="pdf", dpi=300, transparent=True)


# ===========plotting functions===========
def sph_plot_diracs_plotly(
    colatitude_ref=None,
    azimuth_ref=None,
    colatitude=None,
    azimuth=None,
    dirty_img=None,
    azimuth_grid=None,
    colatitude_grid=None,
    surface_base=1,
    surface_height=0.0,
):
    """
    Plots a 2D map on a sphere as well as a collection of diracs using the plotly library

    Parameters
    ----------
    colatitude_ref: ndarray, optional
        The colatitudes of a collection of reference points
    azimuths_ref: ndarray, optional
        The azimuths of a collection of reference points for the Diracs
    colatitude: ndarray, optional
        The colatitudes of the collection of points to visualize
    azimuth: ndarray, optional
        The azimuths of the collection of points to visualize
    dirty_img: ndarray
        A 2D map for displaying a pattern on the sphere under the points
    azimuth_grid: ndarray
        The azimuths indexing the dirty_img 2D map
    colatitude_grid: ndarray
        The colatitudes indexing the dirty_img 2D map
    surface_base:
        radius corresponding to lowest height on the map
    sufrace_height:
        radius difference between the lowest and highest point on the map
    """

    try:
        from plotly.offline import plot
        import plotly.graph_objs as go
        import plotly
    except ImportError:
        import warnings

        warnings.warn("The plotly package is required to use this function")
        return

    plotly.offline.init_notebook_mode()

    traces = []

    if (
        dirty_img is not None
        and azimuth_grid is not None
        and colatitude_grid is not None
    ):

        surfacecolor = np.abs(dirty_img)  # for plotting purposes

        base = surface_base

        surf_diff = surfacecolor.max() - surfacecolor.min()
        if surf_diff > 0:
            height = surface_height / surf_diff
        else:
            height = 0

        r_surf = base + height * (surfacecolor - surfacecolor.min()) / (
            surfacecolor.max() - surfacecolor.min()
        )

        x_plt = r_surf * np.sin(colatitude_grid) * np.cos(azimuth_grid)
        y_plt = r_surf * np.sin(colatitude_grid) * np.sin(azimuth_grid)
        z_plt = r_surf * np.cos(colatitude_grid)

        trace1 = go.Surface(
            x=x_plt,
            y=y_plt,
            z=z_plt,
            surfacecolor=surfacecolor,
            opacity=1,
            colorscale="Portland",
            hoverinfo="none",
        )

        trace1["contours"]["x"]["highlightwidth"] = 1
        trace1["contours"]["y"]["highlightwidth"] = 1
        trace1["contours"]["z"]["highlightwidth"] = 1

        traces.append(trace1)

    if colatitude_ref is not None and azimuth_ref is not None:

        x_ref = np.sin(colatitude_ref) * np.cos(azimuth_ref)
        y_ref = np.sin(colatitude_ref) * np.sin(azimuth_ref)
        z_ref = np.cos(colatitude_ref)

        if not hasattr(colatitude_ref, "__iter__"):
            colatitude_ref = [colatitude_ref]
            azimuth_ref = [azimuth_ref]
            x_ref = [x_ref]
            y_ref = [y_ref]
            z_ref = [z_ref]

        text_str2 = []
        for count, colatitude0 in enumerate(colatitude_ref):
            text_str2.append(
                "({0:.2f}\N{DEGREE SIGN}, {1:.2f}\N{DEGREE SIGN})".format(
                    np.degrees(colatitude0), np.degrees(azimuth_ref[count])
                )
            )

        trace2 = go.Scatter3d(
            mode="markers",
            name="ground truth",
            x=x_ref,
            y=y_ref,
            z=z_ref,
            text=text_str2,
            hoverinfo="name+text",
            marker=dict(
                size=6,
                symbol="circle",
                opacity=0.6,
                line=dict(color="rgb(204, 204, 204)", width=2),
                color="rgb(0, 0.447, 0.741)",
            ),
        )
        traces.append(trace2)

    if colatitude is not None and azimuth is not None:

        x_recon = np.sin(colatitude) * np.cos(azimuth)
        y_recon = np.sin(colatitude) * np.sin(azimuth)
        z_recon = np.cos(colatitude)

        if not hasattr(colatitude, "__iter__"):
            colatitude_ref = [colatitude]
            azimuth = [azimuth]
            x_recon = [x_recon]
            y_recon = [y_recon]
            z_recon = [z_recon]

        text_str3 = []
        for count, colatitude0 in enumerate(colatitude):
            text_str3.append(
                "({0:.2f}\N{DEGREE SIGN}, {1:.2f}\N{DEGREE SIGN})".format(
                    np.degrees(colatitude0), np.degrees(azimuth[count])
                )
            )

        trace3 = go.Scatter3d(
            mode="markers",
            name="reconstruction",
            x=x_recon,
            y=y_recon,
            z=z_recon,
            text=text_str3,
            hoverinfo="name+text",
            marker=dict(
                size=6,
                symbol="diamond",
                opacity=0.6,
                line=dict(color="rgb(204, 204, 204)", width=2),
                color="rgb(0.850, 0.325, 0.098)",
            ),
        )
        traces.append(trace3)

    data = go.Data(traces)

    layout = go.Layout(
        title="",
        autosize=False,
        width=670,
        height=550,
        showlegend=True,
        margin=go.Margin(l=45, r=45, b=55, t=45),
    )

    layout["legend"]["xanchor"] = "center"
    layout["legend"]["yanchor"] = "top"
    layout["legend"]["x"] = 0.5

    fig = go.Figure(data=data, layout=layout)
    plot(fig)


def sph_plot_diracs(
    colatitude_ref=None,
    azimuth_ref=None,
    colatitude=None,
    azimuth=None,
    dirty_img=None,
    colatitude_grid=None,
    azimuth_grid=None,
    file_name="sph_recon_2d_dirac.pdf",
    **kwargs
):
    """
    This function plots the dirty image with sources locations on
    a flat projection of the sphere

    Parameters
    ----------
    colatitude_ref: ndarray, optional
        The colatitudes of a collection of reference points
    azimuths_ref: ndarray, optional
        The azimuths of a collection of reference points for the Diracs
    colatitude: ndarray, optional
        The colatitudes of the collection of points to visualize
    azimuth: ndarray, optional
        The azimuths of the collection of points to visualize
    dirty_img: ndarray
        A 2D map for displaying a pattern on the sphere under the points
    azimuth_grid: ndarray
        The azimuths indexing the dirty_img 2D map
    colatitude_grid: ndarray
        The colatitudes indexing the dirty_img 2D map
    """

    try:
        import matplotlib.pyplot as plt
    except ImportError:
        import warnings

        warnings.warn("Matplotlib is required for plotting")
        return

    fig = plt.figure(figsize=(6.47, 4), dpi=90)
    ax = fig.add_subplot(111, projection="mollweide")

    if (
        dirty_img is not None
        and colatitude_grid is not None
        and azimuth_grid is not None
    ):

        azimuth_plt_internal = azimuth_grid.copy()
        azimuth_plt_internal[azimuth_grid > np.pi] -= np.pi * 2
        p_hd = ax.pcolormesh(
            azimuth_plt_internal,
            np.pi / 2.0 - colatitude_grid,
            np.real(dirty_img),
            cmap="Spectral_r",
            linewidth=0,
            alpha=0.3,  # shading='gouraud',
            antialiased=True,
            zorder=0,
        )

        p_hd.set_edgecolor("None")
        p_hdc = fig.colorbar(
            p_hd,
            orientation="horizontal",
            use_gridspec=False,
            anchor=(0.5, 1.2),
            shrink=0.75,
            spacing="proportional",
        )
        # p_hdc.formatter.set_powerlimits((0, 0))
        p_hdc.ax.tick_params(labelsize=8.5)
        p_hdc.update_ticks()

    if colatitude_ref is not None and azimuth_ref is not None:

        if hasattr(colatitude_ref, "__iter__"):
            K = colatitude_ref.size
            x_ref = azimuth_ref.copy()
            y_ref = (
                np.pi * 0.5 - colatitude_ref.copy()
            )  # convert CO-LATITUDE to LATITUDE

            ind = x_ref > np.pi
            x_ref[ind] -= 2 * np.pi
        else:
            K = 1
            x_ref = azimuth_ref
            y_ref = np.pi * 0.5 - colatitude_ref

            if x_ref > np.pi:
                x_ref = x_ref - 2 * np.pi

        ax.scatter(
            x_ref,
            y_ref,
            c=np.tile([0, 0.447, 0.741], (K, 1)),
            s=70,
            alpha=0.75,
            zorder=11,
            marker="^",
            linewidths=0,
            cmap="Spectral_r",
            label="Original",
        )

    if colatitude is not None and azimuth is not None:

        if hasattr(colatitude, "__iter__"):
            K_est = colatitude.size
            x = azimuth.copy()
            y = np.pi * 0.5 - colatitude.copy()  # convert CO-LATITUDE to LATITUDE

            ind = x > np.pi
            x[ind] -= 2 * np.pi  # scale conversion to -pi to pi
        else:
            K_est = 1
            x = azimuth
            y = np.pi * 0.5 - colatitude

            if x > np.pi:
                x = x - 2 * np.pi

        ax.scatter(
            x,
            y,
            c=np.tile([0.850, 0.325, 0.098], (K_est, 1)),
            s=100,
            alpha=0.75,
            zorder=12,
            marker="*",
            linewidths=0,
            cmap="Spectral_r",
            label="Reconstruction",
        )

    ax.set_xticklabels(
        [
            "210\N{DEGREE SIGN}",
            "240\N{DEGREE SIGN}",
            "270\N{DEGREE SIGN}",
            "300\N{DEGREE SIGN}",
            "330\N{DEGREE SIGN}",
            "0\N{DEGREE SIGN}",
            "30\N{DEGREE SIGN}",
            "60\N{DEGREE SIGN}",
            "90\N{DEGREE SIGN}",
            "120\N{DEGREE SIGN}",
            "150\N{DEGREE SIGN}",
        ]
    )

    ax.legend(
        scatterpoints=1,
        loc=8,
        fontsize=9,
        ncol=2,
        bbox_to_anchor=(0.5, -0.18),
        handletextpad=0.2,
        columnspacing=1.7,
        labelspacing=0.1,
    )  # framealpha=0.3,

    if "title_str" in kwargs:
        ax.set_title(title_str, fontsize=11)

    ax.set_xlabel(r"azimuth", fontsize=11)
    ax.set_ylabel(r"latitude", fontsize=11)

    ax.xaxis.set_label_coords(0.5, 0.52)

    ax.grid(True)

    """
    if save_fig:
        plt.savefig(file_name, format='pdf', dpi=300, transparent=True)
    """



import pyroomacoustics as pra
import numpy as np
from unittest import TestCase

import pyroomacoustics.directivities


class TestSourceDirectivityFlipping(TestCase):
    def test_x(self):
        # create room
        mic_loc = [5, 12, 5]
        source_loc = [5, 2, 5]
        room = (
            pra.ShoeBox(
                p=[10, 14, 10],
                max_order=1,
            )
            .add_source(source_loc)
            .add_microphone(mic_loc)
        )

        # compute image sources
        room.image_source_model()

        # compute azimuth_s and colatitude_s pair for images along x-axis
        source_angle_array = pyroomacoustics.directivities.source_angle_shoebox(
            image_source_loc=room.sources[0].images,
            wall_flips=abs(room.sources[0].orders_xyz),
            mic_loc=mic_loc,
        )
        source_angle_array = np.array(source_angle_array)

        x1_idx = np.where(
            room.sources[0].images[0] == source_loc[0] - 2 * source_loc[0]
        )[0][0]
        x2_idx = np.where(
            room.sources[0].images[0] == source_loc[0] + 2 * source_loc[0]
        )[0][0]
        np.testing.assert_almost_equal(
            source_angle_array[:, x1_idx], [3 * np.pi / 4, np.pi / 2]
        )
        np.testing.assert_almost_equal(
            source_angle_array[:, x2_idx], [np.pi / 4, np.pi / 2]
        )

    def test_y(self):
        # create room
        mic_loc = [12, 5, 5]
        source_loc = [2, 5, 5]
        room = (
            pra.ShoeBox(
                p=[14, 10, 10],
                max_order=1,
            )
            .add_source(source_loc)
            .add_microphone(mic_loc)
        )

        # compute image sources
        room.image_source_model()

        # compute azimuth_s and colatitude_s pair for images along x-axis
        source_angle_array = pyroomacoustics.directivities.source_angle_shoebox(
            image_source_loc=room.sources[0].images,
            wall_flips=abs(room.sources[0].orders_xyz),
            mic_loc=mic_loc,
        )
        source_angle_array = np.array(source_angle_array)

        y1_idx = np.where(
            room.sources[0].images[1] == source_loc[1] - 2 * source_loc[1]
        )[0][0]
        y2_idx = np.where(
            room.sources[0].images[1] == source_loc[1] + 2 * source_loc[1]
        )[0][0]
        np.testing.assert_almost_equal(
            source_angle_array[:, y1_idx], [-np.pi / 4, np.pi / 2]
        )
        np.testing.assert_almost_equal(
            source_angle_array[:, y2_idx], [np.pi / 4, np.pi / 2]
        )

    def test_z(self):
        # create room
        mic_loc = [12, 5, 5]
        source_loc = [2, 5, 5]
        room = (
            pra.ShoeBox(p=[14, 10, 10], max_order=1)
            .add_source(source_loc)
            .add_microphone([12, 5, 5])
        )

        # compute image sources
        room.image_source_model()

        # compute azimuth_s and colatitude_s pair for images along z-axis
        source_angle_array = pyroomacoustics.directivities.source_angle_shoebox(
            image_source_loc=room.sources[0].images,
            wall_flips=abs(room.sources[0].orders_xyz),
            mic_loc=mic_loc,
        )
        source_angle_array = np.array(source_angle_array)

        z1_idx = np.where(
            room.sources[0].images[2] == source_loc[2] - 2 * source_loc[2]
        )[0][0]
        z2_idx = np.where(
            room.sources[0].images[2] == source_loc[2] + 2 * source_loc[2]
        )[0][0]
        np.testing.assert_almost_equal(
            source_angle_array[:, z1_idx], [0, 3 * np.pi / 4]
        )
        np.testing.assert_almost_equal(source_angle_array[:, z2_idx], [0, np.pi / 4])



# Author: Eric Bezzam
# Date: July 15, 2016
from __future__ import division, print_function

from .doa import *


class SRP(DOA):
    """
    Class to apply Steered Response Power (SRP) direction-of-arrival (DoA) for
    a particular microphone array.

    .. note:: Run locate_source() to apply the SRP-PHAT algorithm.

    Parameters
    ----------
    L: numpy array
        Microphone array positions. Each column should correspond to the
        cartesian coordinates of a single microphone.
    fs: float
        Sampling frequency.
    nfft: int
        FFT length.
    c: float
        Speed of sound. Default: 343 m/s
    num_src: int
        Number of sources to detect. Default: 1
    mode: str
        'far' or 'near' for far-field or near-field detection
        respectively. Default: 'far'
    r: numpy array
        Candidate distances from the origin. Default: np.ones(1)
    azimuth: numpy array
        Candidate azimuth angles (in radians) with respect to x-axis.
        Default: np.linspace(-180.,180.,30)*np.pi/180
    colatitude: numpy array
        Candidate elevation angles (in radians) with respect to z-axis.
        Default is x-y plane search: np.pi/2*np.ones(1)
    """

    def __init__(
        self,
        L,
        fs,
        nfft,
        c=343.0,
        num_src=1,
        mode="far",
        r=None,
        azimuth=None,
        colatitude=None,
        **kwargs
    ):

        DOA.__init__(
            self,
            L=L,
            fs=fs,
            nfft=nfft,
            c=c,
            num_src=num_src,
            mode=mode,
            r=r,
            azimuth=azimuth,
            colatitude=colatitude,
            **kwargs
        )

        self.num_pairs = self.M * (self.M - 1) / 2

    def _process(self, X):
        """
        Perform SRP-PHAT for given frame in order to estimate steered response
        spectrum.
        """

        srp_cost = np.zeros(self.grid.n_points)

        # apply PHAT weighting
        absX = np.abs(X)
        absX[absX < tol] = tol
        pX = X / absX

        CC = []
        for k in self.freq_bins:
            CC.append(np.dot(pX[:, k, :], np.conj(pX[:, k, :]).T))
        CC = np.array(CC)

        M = self.L.shape[1]
        ar = np.arange(M)
        av = ar[:, None]

        # the mask here allow to select all the coefficients above
        # the main diagonal of the covariance matrix
        # It can be applied on the flattened last two dimensions of the
        # stack of cov. matrices
        mask_triu = (av < av.T).flatten()

        # Flatten the covariance matrices and use the above mask to
        # select the upper triangular part
        CC_flat = CC.reshape((-1, CC.shape[-2] * CC.shape[-2]))[:, mask_triu]

        # The DC offset is the sum of all the diagonal coefficients
        # Due to the normalization in SRP-PHAT, they are all ones,
        # and we end up with the product:
        # <number of frames> x <number of microphones> x <number of freqs>
        # the number of frames appears because the covariance matrix
        # is not normalized
        DC_offset = pX.shape[-1] * self.L.shape[1] * len(self.freq_bins)

        for n in range(self.grid.n_points):

            # In the loop, this is just a fancy way of computing
            # the quadratic form:
            # mode_vec^H @ CC @ mode_vec

            # get the mode vector axis: (frequency, microphones)
            mode_vec = self.mode_vec[self.freq_bins, :, n]

            # compute the outer product along the microphone axis
            mode_mat = np.conj(mode_vec[:, :, None]) * mode_vec[:, None, :]

            # First, we flatten the mode vector covariance and select the
            # terms above the main diagonal
            mode_mat_flat = mode_mat.reshape((-1, M * M))[:, mask_triu]
            # Then we compute manually the real part of the element-wise product
            # This is equivalent to `np.real(CC_flat * mode_mat_flat)
            # but avoids computing the imaginary part that we end up discarding
            R = CC_flat.real * mode_mat_flat.real - CC_flat.imag * mode_mat_flat.imag
            # Finally, we sum up and add the DC offset to make the cost non-negative
            sum_val = 2.0 * np.sum(R) + DC_offset

            # Finally normalize
            srp_cost[n] = sum_val / self.num_snap / self.num_freq / self.num_pairs

        self.grid.set_values(srp_cost)



# @version: 1.0  date: 2017/06/20 by Robin Scheibler
# @author: robin.scheibler@epfl.ch, ivan.dokmanic@epfl.ch, sidney.barthe@epfl.ch
# @copyright: EPFL-IC-LCAV 2017

import unittest

import numpy as np
import pyroomacoustics as pra
from scipy.signal import fftconvolve

# fix the RNG seed for repeatability
np.random.seed(0)

# Location of original source
azimuth = 61.0 / 180.0 * np.pi  # 60 degrees
tol = 0.3 / 180.0 * np.pi  # 0.3 degrees tolerance for the test

# algorithms parameters
c = 343.0
fs = 16000
nfft = 256
freq_bins = np.arange(5, 60)

# circular microphone array, 12 mics, radius 15 cm
R = pra.circular_2D_array([0, 0], 12, 0.0, 0.15)

# propagation filter bank
propagation_vector = -np.array([np.cos(azimuth), np.sin(azimuth)])
delays = np.dot(R.T, propagation_vector) / c * fs  # in fractional samples
filter_bank = pra.fractional_delay_filter_bank(delays)

# we use a white noise signal for the source
x = np.random.randn((nfft // 2 + 1) * nfft)

# convolve the source signal with the fractional delay filters
# to get the microphone input signals
mic_signals = np.array([fftconvolve(x, filter, mode="same") for filter in filter_bank])
X = pra.transform.stft.analysis(mic_signals.T, nfft, nfft // 2, win=np.hanning(nfft))
X = np.swapaxes(X, 2, 0)


class TestDOA(unittest.TestCase):
    def test_music(self):
        doa = pra.doa.algorithms["MUSIC"](R, fs, nfft, c=c)
        doa.locate_sources(X, freq_bins=freq_bins)
        print("distance:", pra.doa.circ_dist(azimuth, doa.azimuth_recon))
        self.assertTrue(pra.doa.circ_dist(azimuth, doa.azimuth_recon) < tol)

    def test_normmusic(self):
        doa = pra.doa.algorithms["NormMUSIC"](R, fs, nfft, c=c)
        doa.locate_sources(X, freq_bins=freq_bins)
        print("distance:", pra.doa.circ_dist(azimuth, doa.azimuth_recon))
        self.assertTrue(pra.doa.circ_dist(azimuth, doa.azimuth_recon) < tol)

    def test_srp_phat(self):
        doa = pra.doa.algorithms["SRP"](R, fs, nfft, c=c)
        doa.locate_sources(X, freq_bins=freq_bins)
        print("distance:", pra.doa.circ_dist(azimuth, doa.azimuth_recon))
        self.assertTrue(pra.doa.circ_dist(azimuth, doa.azimuth_recon) < tol)

    def test_srp_phat_2ch(self):
        """
        Tests that the cost function of SRP is properly computed (PR #197)

        This checks for the presence of a bug that was detected in PR #197.
        """
        doa = pra.doa.algorithms["SRP"](R[:, :2], fs, nfft, c=c)
        i_freq = 10
        doa.locate_sources(X[:2, :, :], freq_bins=freq_bins[i_freq : i_freq + 1])
        std_val = np.std(doa.grid.values)
        self.assertTrue(std_val > 1e-10)

    def test_cssm(self):
        doa = pra.doa.algorithms["CSSM"](R, fs, nfft, c=c)
        doa.locate_sources(X, freq_bins=freq_bins)
        print("distance:", pra.doa.circ_dist(azimuth, doa.azimuth_recon))
        self.assertTrue(pra.doa.circ_dist(azimuth, doa.azimuth_recon) < tol)

    def test_tops(self):
        doa = pra.doa.algorithms["TOPS"](R, fs, nfft, c=c)
        doa.locate_sources(X, freq_bins=freq_bins)
        print("distance:", pra.doa.circ_dist(azimuth, doa.azimuth_recon))
        self.assertTrue(pra.doa.circ_dist(azimuth, doa.azimuth_recon) < tol)

    def test_waves(self):
        doa = pra.doa.algorithms["WAVES"](R, fs, nfft, c=c)
        doa.locate_sources(X, freq_bins=freq_bins)
        print("distance:", pra.doa.circ_dist(azimuth, doa.azimuth_recon))
        self.assertTrue(pra.doa.circ_dist(azimuth, doa.azimuth_recon) < tol)

    def test_frida(self):
        doa = pra.doa.algorithms["FRIDA"](R, fs, nfft, c=c)
        doa.locate_sources(X, freq_bins=freq_bins)
        print("distance:", pra.doa.circ_dist(azimuth, doa.azimuth_recon))
        self.assertTrue(pra.doa.circ_dist(azimuth, doa.azimuth_recon) < tol)


if __name__ == "__main__":

    """
    algo_names = sorted(pra.doa.algorithms.keys())

    for algo_name in algo_names:
        doa = pra.doa.algorithms[algo_name](R, fs, nfft, c=c, max_four=4)
        doa.locate_sources(X, freq_bins=freq_bins)
        print(
            algo_name,
            doa.azimuth_recon / np.pi * 180.0,
            pra.doa.circ_dist(azimuth, doa.azimuth_recon) / np.pi * 180.0,
        )
    """

    unittest.main()



# Test of the CCW3P routine
# Copyright (C) 2019  Robin Scheibler, Cyril Cadoux
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
# You should have received a copy of the MIT License along with this program. If
# not, see <https://opensource.org/licenses/MIT>.

import numpy as np
import pyroomacoustics as pra

cases = {
    "anti-clockwise": {
        "points": np.array([[1, -1], [2, -1], [1, 0]]),
        "expected": 1,  # anti-clockwise
        "label": "Test: CCW3P anti-clockwise",
    },
    "clockwise": {
        "points": np.array([[1, -1], [1, 0], [2, -1]]),
        "expected": -1,  # clockwise
        "label": "Test: CCW3P clockwise",
    },
    "co-linear": {
        "points": np.array([[0, 0], [0.5, 0.5], [1, 1]]),
        "expected": 0,  # co-linear
        "label": "Test: CCW3P co-linear",
    },
}


def ccw3p(case):

    p1, p2, p3 = case["points"]

    r = pra.libroom.ccw3p(p1, p2, p3)

    assert r == case["expected"], case["label"] + " returned: {}, expected {}".format(
        r, case["expected"]
    )


def test_ccw3p_anticlockwise():
    ccw3p(cases["anti-clockwise"])


def test_ccw3p_clockwise():
    ccw3p(cases["clockwise"])


def test_ccw3p_colinear():
    ccw3p(cases["co-linear"])


if __name__ == "__main__":

    for lbl, case in cases.items():

        try:
            ccw3p(case)
        except:
            print("{} failed".format(lbl))



import numpy as np
from scipy.signal import fftconvolve


def trinicon(
    signals,
    w0=None,
    filter_length=2048,
    block_length=None,
    n_blocks=8,
    alpha_on=None,
    j_max=10,
    delta_max=1e-4,
    sigma2_0=1e-7,
    mu=0.001,
    lambd_a=0.2,
    return_filters=False,
):
    """
    Implementation of the TRINICON Blind Source Separation algorithm as described in

    R. Aichner, H. Buchner, F. Yan, and W. Kellermann  *A real-time
    blind source separation scheme and its application to reverberant and noisy
    acoustic environments*,  Signal Processing, 86(6), 1260-1277.
    doi:10.1016/j.sigpro.2005.06.022, 2006. `[pdf] <http://www.buchner-net.com/lnt2006_19.pdf>`_

    Specifically, adaptation of the pseudo-code from Table 1.

    The implementation is hard-coded for 2 output channels.

    Parameters
    ----------
    signals: ndarray (nchannels, nsamples)
        The microphone input signals (time domain)
    w0: ndarray (nchannels, nsources, nsamples), optional
        Optional initial value for the demixing filters
    filter_length: int, optional
        The length of the demixing filters, if w0 is provided, this option is ignored
    block_length: int, optional
        Block length (default 2x filter_length)
    n_blocks: int, optional
        Number of blocks processed at once (default 8)
    alpha_on: int, optional
        Online overlap factor (default ``n_blocks``)
    j_max: int, optional
        Number of offline iterations (default 10)
    delta_max: float, optional
        Regularization parameter, this sets the maximum value of the regularization term (default 1e-4)
    sigma2_0: float, optional
        Regularization parameter, this sets the reference (machine?) noise
        level in the regularization (default 1e-7)
    mu: float, optional
        Offline update step size (default 0.001)
    lambd_a: float, optional
        Online forgetting factor (default 0.2)
    return_filters: bool
        If true, the function will return the demixing matrix too (default False)

    Returns
    -------
    ndarray
        Returns an (nsources, nsamples) array. Also returns
        the demixing matrix (nchannels, nsources, nsamples)
        if ``return_filters`` keyword is True.
    """

    P = signals.shape[0]  # number of microphones
    Q = 2  # number of output channels

    # the filters
    if w0 is None:
        L = filter_length
        w = np.zeros((P, Q, L))
        w[: P // 2, 0, L // 2] = 1
        w[P // 2 :, 1, L // 2] = 1
    else:
        w = w0.copy()
        L = w0.shape[2]

    K = n_blocks  # Number of successive blocks processed at the same time
    if block_length is None:  # Block length
        N = 2 * L
    else:
        N = block_length

    if alpha_on is None:
        alpha_on = K  # online overlap factor

    hop = K * L // alpha_on

    # pad with zeros to have a whole number of online blocks
    if signals.shape[1] % hop != 0:
        signals = np.concatenate(
            (signals, np.zeros((P, hop - (signals.shape[0] % hop)))), axis=1
        )

    S = signals.shape[1]  # total signal length
    M = S / hop  # number of online blocks

    y = np.zeros((Q, S))  # the processed output signal

    m = 1  # online block index
    while m <= M:  # online loop

        # new chunk of input signal
        x = np.zeros((P, K * L + N))
        if m * hop > S:
            # we need some zero padding at the back
            le = S - (m - 1) * hop + N
            x[:le] = signals[:, m * hop - K * L - N :]
        if m * hop >= K * L + N:
            x = signals[:, m * hop - K * L - N : m * hop]
        else:
            # we need some zero padding at the beginning
            x[:, -m * hop :] = signals[:, : m * hop]

        # use filter from previous iteration to initialize offline part
        w_new = w.copy()

        for j in range(j_max):  # offline update loop

            y_c = np.zeros((Q, K * L + N - L))  # c stands for chunk
            y_blocks = np.zeros((Q, K, N))

            for q in range(Q):
                # convolve with filters
                for p in range(P):
                    # We discard the 'oldest' output of the convolution according
                    # to the filter matrix definition (6) in the paper
                    y_c[q, :] += fftconvolve(x[p, :], w_new[p, q, :], mode="valid")[1:]

                # split into smaller blocks
                for i in range(K):
                    y_blocks[q, i, :] = y_c[q, i * L : i * L + N]

            # blocks energy
            sigma2 = np.sum(y_blocks**2, axis=2)

            # cross-correlations
            # XXX This is the part hard coded for two channels XXX
            r_cross = np.zeros((Q, K, 2 * L - 1))
            for i in range(K):
                y0 = y_c[0, i * L : i * L + N]
                y1 = y_c[1, i * L : i * L + N]
                r = fftconvolve(y1, y0[::-1], mode="full")
                r_cross[0, i, :] = r[N - L : N + L - 1]  # r_y1y0
                r_cross[1, i, :] = r_cross[
                    0, i, ::-1
                ]  # r_y0y1 by symmetry is just r_y1y0 reversed

            # regularization term
            delta = delta_max * np.exp(-sigma2 / sigma2_0)

            # offline update
            delta_w = np.zeros((P, Q, L))
            for q in range(Q):
                for p in range(P):
                    for i in range(K):
                        # this implements the row-wise sylvester constraint as explained in Fig. 4 (b) of paper
                        delta_w[p, q, :] += fftconvolve(
                            r_cross[q, i, :] / (sigma2[q, i] + delta[q, i]),
                            w_new[p, 1 - q, ::-1],
                            mode="valid",
                        )[::-1]
                    delta_w[p, q, :] /= K

            w_new = w_new - mu * delta_w

        # online update
        w = lambd_a * w + (1 - lambd_a) * w_new

        # compute output signal
        for q in range(Q):
            for p in range(P):
                y[q, (m - 1) * hop : m * hop] += fftconvolve(
                    x[p, -hop - L + 1 :], w[p, q, :], mode="valid"
                )

        # next block
        m += 1

    if return_filters:
        return y, w
    else:
        return y



import abc
import numpy as np
from enum import Enum
from pyroomacoustics.doa import spher2cart
from pyroomacoustics.utilities import requires_matplotlib, all_combinations


class DirectivityPattern(Enum):
    """
    Common cardioid patterns and their corresponding coefficient for the expression:

    .. math::

        r = a (1 + \\cos \\theta),

    where :math:`a` is the coefficient that determines the cardioid pattern and :math:`r` yields
    the gain at angle :math:`\\theta`.

    """

    FIGURE_EIGHT = 0
    HYPERCARDIOID = 0.25
    CARDIOID = 0.5
    SUBCARDIOID = 0.75
    OMNI = 1.0


class DirectionVector(object):
    """
    Object for representing direction vectors in 3D, parameterized by an azimuth and colatitude
    angle.

    Parameters
    ----------
    azimuth : float
    colatitude : float, optional
        Default to PI / 2, only XY plane.
    degrees : bool
        Whether provided values are in degrees (True) or radians (False).
    """

    def __init__(self, azimuth, colatitude=None, degrees=True):
        if degrees is True:
            azimuth = np.radians(azimuth)
            if colatitude is not None:
                colatitude = np.radians(colatitude)
        self._azimuth = azimuth
        if colatitude is None:
            colatitude = np.pi / 2
        assert colatitude <= np.pi and colatitude >= 0
        self._colatitude = colatitude

        self._unit_v = np.array(
            [
                np.cos(self._azimuth) * np.sin(self._colatitude),
                np.sin(self._azimuth) * np.sin(self._colatitude),
                np.cos(self._colatitude),
            ]
        )

    def get_azimuth(self, degrees=False):
        if degrees:
            return np.degrees(self._azimuth)
        else:
            return self._azimuth

    def get_colatitude(self, degrees=False):
        if degrees:
            return np.degrees(self._colatitude)
        else:
            return self._colatitude

    @property
    def unit_vector(self):
        """Direction vector in cartesian coordinates."""
        return self._unit_v


class Directivity(abc.ABC):
    """
    Abstract class for directivity patterns.

    """

    def __init__(self, orientation):
        assert isinstance(orientation, DirectionVector)
        self._orientation = orientation

    def get_azimuth(self, degrees=True):
        return self._orientation.get_azimuth(degrees)

    def get_colatitude(self, degrees=True):
        return self._orientation.get_colatitude(degrees)

    def set_orientation(self, orientation):
        """
        Set orientation of directivity pattern.

        Parameters
        ----------
        orientation : DirectionVector
            New direction for the directivity pattern.
        """
        assert isinstance(orientation, DirectionVector)
        self._orientation = orientation

    @abc.abstractmethod
    def get_response(
        self, azimuth, colatitude=None, magnitude=False, frequency=None, degrees=True
    ):
        """
        Get response for provided angles and frequency.
        """
        return


class CardioidFamily(Directivity):
    """
    Object for directivities coming from the
    `cardioid family <https://en.wikipedia.org/wiki/Microphone#Cardioid,_hypercardioid,_supercardioid,_subcardioid>`_.

    Parameters
    ----------
    orientation : DirectionVector
        Indicates direction of the pattern.
    pattern_enum : DirectivityPattern
        Desired pattern for the cardioid.
    """

    def __init__(self, orientation, pattern_enum, gain=1.0):
        Directivity.__init__(self, orientation)
        self._p = pattern_enum.value
        self._gain = gain
        self._pattern_name = pattern_enum.name

    @property
    def directivity_pattern(self):
        """Name of cardioid directivity pattern."""
        return self._pattern_name

    def get_response(
        self, azimuth, colatitude=None, magnitude=False, frequency=None, degrees=True
    ):
        """
        Get response for provided angles.

        Parameters
        ----------
        azimuth : array_like
            Azimuth in degrees
        colatitude : array_like, optional
            Colatitude in degrees. Default is to be on XY plane.
        magnitude : bool, optional
            Whether to return magnitude of response.
        frequency : float, optional
            For which frequency to compute the response. Cardioid are frequency-independent so this
            value has no effect.
        degrees : bool, optional
            Whether provided angles are in degrees.


        Returns
        -------
        resp : :py:class:`~numpy.ndarray`
            Response at provided angles.
        """

        if colatitude is not None:
            assert len(azimuth) == len(colatitude)
        if self._p == DirectivityPattern.OMNI:
            return np.ones(len(azimuth))
        else:
            coord = spher2cart(azimuth=azimuth, colatitude=colatitude, degrees=degrees)
            resp = self._gain * self._p + (1 - self._p) * np.matmul(
                self._orientation.unit_vector, coord
            )
            if magnitude:
                return np.abs(resp)
            else:
                return resp

    @requires_matplotlib
    def plot_response(
        self, azimuth, colatitude=None, degrees=True, ax=None, offset=None
    ):
        """
        Plot directivity response at specified angles.

        Parameters
        ----------
        azimuth : array_like
            Azimuth values for plotting.
        colatitude : array_like, optional
            Colatitude values for plotting. If not provided, 2D plot.
        degrees : bool
            Whether provided values are in degrees (True) or radians (False).
        ax : axes object
        offset : list
            3-D coordinates of the point where the response needs to be plotted.

        Return
        ------
        ax : :py:class:`~matplotlib.axes.Axes`
        """
        import matplotlib.pyplot as plt

        if offset is not None:
            x_offset = offset[0]
            y_offset = offset[1]
        else:
            x_offset = 0
            y_offset = 0

        if degrees:
            azimuth = np.radians(azimuth)

        if colatitude is not None:

            if degrees:
                colatitude = np.radians(colatitude)

            if ax is None:
                fig = plt.figure()
                ax = fig.add_subplot(1, 1, 1, projection="3d")

            if offset is not None:
                z_offset = offset[2]
            else:
                z_offset = 0

            spher_coord = all_combinations(azimuth, colatitude)
            azi_flat = spher_coord[:, 0]
            col_flat = spher_coord[:, 1]

            # compute response
            resp = self.get_response(
                azimuth=azi_flat, colatitude=col_flat, magnitude=True, degrees=False
            )
            RESP = resp.reshape(len(azimuth), len(colatitude))

            # create surface plot, need cartesian coordinates
            AZI, COL = np.meshgrid(azimuth, colatitude)
            X = RESP.T * np.sin(COL) * np.cos(AZI) + x_offset
            Y = RESP.T * np.sin(COL) * np.sin(AZI) + y_offset
            Z = RESP.T * np.cos(COL) + z_offset

            ax.plot_surface(X, Y, Z)

            if ax is None:
                ax.set_title(
                    "{}, azimuth={}, colatitude={}".format(
                        self.directivity_pattern,
                        self.get_azimuth(),
                        self.get_colatitude(),
                    )
                )
            else:
                ax.set_title("Directivity Plot")

            ax.set_xlabel("x")
            ax.set_ylabel("y")
            ax.set_zlabel("z")

        else:

            if ax is None:
                fig = plt.figure()
                ax = plt.subplot(111)

            # compute response
            resp = self.get_response(azimuth=azimuth, magnitude=True, degrees=False)
            RESP = resp

            # create surface plot, need cartesian coordinates
            X = RESP.T * np.cos(azimuth) + x_offset
            Y = RESP.T * np.sin(azimuth) + y_offset
            ax.plot(X, Y)

        return ax


def cardioid_func(x, direction, coef, gain=1.0, normalize=True, magnitude=False):
    """
    One-shot function for computing cardioid response.

    Parameters
    -----------
    x: array_like, shape (..., n_dim)
         Cartesian coordinates
    direction: array_like, shape (n_dim)
         Direction vector, should be normalized.
    coef: float
         Parameter for the cardioid function.
    gain: float
         The gain.
    normalize : bool
        Whether to normalize coordinates and direction vector.
    magnitude : bool
        Whether to return magnitude, default is False.

    Returns
    -------
    resp : :py:class:`~numpy.ndarray`
        Response at provided angles for the speficied cardioid function.
    """
    assert coef >= 0.0
    assert coef <= 1.0

    # normalize positions
    if normalize:
        x /= np.linalg.norm(x, axis=0)
        direction /= np.linalg.norm(direction)

    # compute response
    resp = gain * (coef + (1 - coef) * np.matmul(direction, x))
    if magnitude:
        return np.abs(resp)
    else:
        return resp


def source_angle_shoebox(image_source_loc, wall_flips, mic_loc):
    """
    Determine outgoing angle for each image source for a ShoeBox configuration.

    Implementation of the method described in the paper:
    https://www2.ak.tu-berlin.de/~akgroup/ak_pub/2018/000458.pdf

    Parameters
    -----------
    image_source_loc : array_like
        Locations of image sources.
    wall_flips: array_like
        Number of x, y, z flips for each image source.
    mic_loc: array_like
        Microphone location.

    Returns
    -------
    azimuth : :py:class:`~numpy.ndarray`
        Azimith for each image source, in radians
    colatitude : :py:class:`~numpy.ndarray`
        Colatitude for each image source, in radians.

    """

    image_source_loc = np.array(image_source_loc)
    wall_flips = np.array(wall_flips)
    mic_loc = np.array(mic_loc)

    dim, n_sources = image_source_loc.shape
    assert wall_flips.shape[0] == dim
    assert mic_loc.shape[0] == dim

    p_vector_array = image_source_loc - np.array(mic_loc)[:, np.newaxis]
    d_array = np.linalg.norm(p_vector_array, axis=0)

    # Using (12) from the paper
    power_array = np.ones_like(image_source_loc) * -1
    power_array = np.power(power_array, (wall_flips + np.ones_like(image_source_loc)))
    p_dash_array = p_vector_array * power_array

    # Using (13) from the paper
    azimuth = np.arctan2(p_dash_array[1], p_dash_array[0])
    if dim == 2:
        colatitude = np.ones(n_sources) * np.pi / 2
    else:
        colatitude = np.pi / 2 - np.arcsin(p_dash_array[2] / d_array)

    return azimuth, colatitude



import pyroomacoustics as pra
import numpy as np
from pyroomacoustics.directivities import (
    DirectivityPattern,
    DirectionVector,
    CardioidFamily,
)
from unittest import TestCase


# create room
room = pra.ShoeBox(
    p=[7, 5, 5],
    materials=pra.Material(0.07),
    fs=16000,
    max_order=1,
)

# define source with figure_eight directivity
PATTERN = DirectivityPattern.FIGURE_EIGHT
ORIENTATION = DirectionVector(azimuth=90, colatitude=90, degrees=True)
directivity = CardioidFamily(orientation=ORIENTATION, pattern_enum=PATTERN)

# add source with figure_eight directivity
room.add_source([2.5, 2.5, 2.5], directivity=directivity)

# add microphone in its null
room.add_microphone([5, 2.5, 2.5])

# compute rir
room.compute_rir()


class TestSourceDirectivity(TestCase):
    def test_zero_gain(self):
        gain = np.amax(room.rir[0][0])
        self.assertTrue(gain < (10**-10))


def get_error():
    gain = np.amax(room.rir[0][0])
    theoretical_gain = 0
    error = abs(gain - theoretical_gain)
    print()
    print("-" * 40)
    print("The error in gain calculation is: {}".format(error))
    print("-" * 40)


if __name__ == "__main__":
    get_error()



# Copyright (c) 2019 Kouhei Sekiguchi, Yoshiaki Bando
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
"""
FastMNMF
========

Blind Source Separation using Fast Multichannel Nonnegative Matrix Factorization (FastMNMF)
"""
import numpy as np


def fastmnmf(
    X,
    n_src=None,
    n_iter=30,
    n_components=8,
    mic_index=0,
    W0=None,
    accelerate=True,
    callback=None,
):
    """
    Implementation of FastMNMF algorithm presented in

    K. Sekiguchi, A. A. Nugraha, Y. Bando, K. Yoshii, *Fast Multichannel Source
    Separation Based on Jointly Diagonalizable Spatial Covariance Matrices*, EUSIPCO, 2019. [`arXiv <https://arxiv.org/abs/1903.03237>`_] [`IEEE <https://ieeexplore.ieee.org/abstract/document/8902557>`_]

    The code of FastMNMF with GPU support and FastMNMF-DP which integrates DNN-based source model
    into FastMNMF is available on https://github.com/sekiguchi92/SoundSourceSeparation

    Parameters
    ----------
    X: ndarray (nframes, nfrequencies, nchannels)
        STFT representation of the observed signal
    n_src: int, optional
        The number of sound sources (default None).
        If None, n_src is set to the number of microphones
    n_iter: int, optional
        The number of iterations (default 30)
    n_components: int, optional
        Number of components in the non-negative spectrum (default 8)
    mic_index: int or 'all', optional
        The index of microphone of which you want to get the source image (default 0).
        If 'all', return the source images of all microphones
    W0: ndarray (nfrequencies, nchannels, nchannels), optional
        Initial value for diagonalizer Q (default None).
        If None, identity matrices are used for all frequency bins.
    accelerate: bool, optional
        If true, the basis and activation of NMF are updated simultaneously (default True)
    callback: func, optional
        A callback function called every 10 iterations, allows to monitor convergence

    Returns
    -------
    If mic_index is int, returns an (nframes, nfrequencies, nsources) array.
    If mic_index is 'all', returns an (nchannels, nframes, nfrequencies, nsources) array.
    """
    eps = 1e-7
    g_eps = 5e-2
    interval_update_Q = 1  # 2 may work as well and is faster
    interval_normalize = 10
    TYPE_FLOAT = X.real.dtype
    TYPE_COMPLEX = X.dtype

    # initialize parameter
    X_FTM = X.transpose(1, 0, 2)
    n_freq, n_frames, n_chan = X_FTM.shape
    XX_FTMM = np.matmul(X_FTM[:, :, :, None], X_FTM[:, :, None, :].conj())
    if n_src is None:
        n_src = X_FTM.shape[2]

    if W0 is not None:
        Q_FMM = W0
    else:
        Q_FMM = np.tile(np.eye(n_chan).astype(TYPE_COMPLEX), [n_freq, 1, 1])

    G_NFM = np.ones([n_src, n_freq, n_chan], dtype=TYPE_FLOAT) * g_eps
    for m in range(n_chan):
        G_NFM[m % n_src, :, m] = 1

    for m in range(n_chan):
        mu_F = (Q_FMM[:, m] * Q_FMM[:, m].conj()).sum(axis=1).real
        Q_FMM[:, m] /= np.sqrt(mu_F[:, None])

    H_NKT = np.random.rand(n_src, n_components, n_frames).astype(TYPE_FLOAT)
    W_NFK = np.random.rand(n_src, n_freq, n_components).astype(TYPE_FLOAT)
    lambda_NFT = W_NFK @ H_NKT + eps
    Qx_power_FTM = np.abs(np.einsum("fij, ftj -> fti", Q_FMM, X_FTM)) ** 2
    Y_FTM = np.einsum("nft, nfm -> ftm", lambda_NFT, G_NFM)

    def separate():
        Qx_FTM = np.einsum("fij, ftj -> fti", Q_FMM, X_FTM)
        Qinv_FMM = np.linalg.inv(Q_FMM)
        Y_NFTM = np.einsum("nft, nfm -> nftm", lambda_NFT, G_NFM)

        if mic_index == "all":
            return np.einsum(
                "fij, ftj, nftj -> itfn", Qinv_FMM, Qx_FTM / Y_NFTM.sum(axis=0), Y_NFTM
            )
        elif type(mic_index) is int:
            return np.einsum(
                "fj, ftj, nftj -> tfn",
                Qinv_FMM[:, mic_index],
                Qx_FTM / Y_NFTM.sum(axis=0),
                Y_NFTM,
            )
        else:
            raise ValueError("mic_index should be int or 'all'")

    # update parameters
    for epoch in range(n_iter):
        if callback is not None and epoch % 10 == 0:
            callback(separate())

        # update W and H (basis and activation of NMF)
        tmp1_NFT = np.einsum("nfm, ftm -> nft", G_NFM, Qx_power_FTM / (Y_FTM**2))
        tmp2_NFT = np.einsum("nfm, ftm -> nft", G_NFM, 1 / Y_FTM)

        numerator = np.einsum("nkt, nft -> nfk", H_NKT, tmp1_NFT)
        denominator = np.einsum("nkt, nft -> nfk", H_NKT, tmp2_NFT) + eps
        W_NFK *= np.sqrt(numerator / denominator)

        if not accelerate:
            lambda_NFT = W_NFK @ H_NKT + eps
            Y_FTM = np.einsum("nft, nfm -> ftm", lambda_NFT, G_NFM)
            tmp1_NFT = np.einsum("nfm, ftm -> nft", G_NFM, Qx_power_FTM / (Y_FTM**2))
            tmp2_NFT = np.einsum("nfm, ftm -> nft", G_NFM, 1 / Y_FTM)

        numerator = np.einsum("nfk, nft -> nkt", W_NFK, tmp1_NFT)
        denominator = np.einsum("nfk, nft -> nkt", W_NFK, tmp2_NFT) + eps
        H_NKT *= np.sqrt(numerator / denominator)

        lambda_NFT = W_NFK @ H_NKT + eps
        Y_FTM = np.einsum("nft, nfm -> ftm", lambda_NFT, G_NFM)

        # update G_NFM (diagonal element of spatial covariance matrices)
        numerator = np.einsum(
            "nft, ftm -> nfm", lambda_NFT, Qx_power_FTM / (Y_FTM**2)
        )
        denominator = np.einsum("nft, ftm -> nfm", lambda_NFT, 1 / Y_FTM) + eps
        G_NFM *= np.sqrt(numerator / denominator)
        Y_FTM = np.einsum("nft, nfm -> ftm", lambda_NFT, G_NFM)

        # udpate Q (matrix for joint diagonalization)
        if (interval_update_Q <= 0) or (epoch % interval_update_Q == 0):
            for m in range(n_chan):
                V_FMM = (
                    np.einsum("ftij, ft -> fij", XX_FTMM, 1 / Y_FTM[..., m]) / n_frames
                )
                tmp_FM = np.linalg.solve(
                    np.matmul(Q_FMM, V_FMM), np.eye(n_chan)[None, m]
                )
                Q_FMM[:, m] = (
                    tmp_FM
                    / (
                        np.sqrt(
                            np.einsum("fi, fij, fj -> f", tmp_FM.conj(), V_FMM, tmp_FM)
                        )[:, None]
                        + eps
                    )
                ).conj()
                Qx_power_FTM = np.abs(np.einsum("fij, ftj -> fti", Q_FMM, X_FTM)) ** 2

        # normalize
        if (interval_normalize <= 0) or (epoch % interval_normalize == 0):
            phi_F = np.einsum("fij, fij -> f", Q_FMM, Q_FMM.conj()).real / n_chan
            Q_FMM /= np.sqrt(phi_F)[:, None, None]
            G_NFM /= phi_F[None, :, None]

            mu_NF = G_NFM.sum(axis=2)
            G_NFM /= mu_NF[..., None]
            W_NFK *= mu_NF[..., None]

            nu_NK = W_NFK.sum(axis=1)
            W_NFK /= nu_NK[:, None]
            H_NKT *= nu_NK[:, :, None]

            lambda_NFT = W_NFK @ H_NKT + eps
            Qx_power_FTM = np.abs(np.einsum("fij, ftj -> fti", Q_FMM, X_FTM)) ** 2
            Y_FTM = np.einsum("nft, nfm -> ftm", lambda_NFT, G_NFM)

    return separate()



import numpy as np
import pyroomacoustics as pra


def test_cart_spher_convertion(repetitions=10, vectorized=True):
    """Tests that the convertion is invertible"""

    np.random.seed(124584)

    for epoch in range(repetitions):
        if vectorized:
            cart = np.random.randn(3, 10)
        else:
            cart = np.random.randn(3)

        sph = np.array(pra.doa.cart2spher(cart))
        cart_back = pra.doa.spher2cart(*sph)
        sph_back = np.array(pra.doa.cart2spher(cart_back))

        err1 = np.max(np.abs(sph_back - sph))
        err2 = np.max(np.abs(cart_back - cart))

        assert err1 < 1e-13
        assert err2 < 1e-13


if __name__ == "__main__":
    test_cart_spher_convertion(vectorized=True)
    test_cart_spher_convertion(vectorized=False)



"""
Point Clouds
============
Contains PointCloud class.

Given a number of points and their relative distances, this class aims at
reconstructing their relative coordinates.
"""

from __future__ import division, print_function

# Provided by LCAV
import numpy as np
from scipy import linalg as la


class PointCloud:
    def __init__(self, m=1, dim=3, diameter=0.0, X=None, labels=None, EDM=None):
        """
        Parameters
        ----------

        m : int, optional
            Number of markers
        diameter : float, optional
            Diameter of the marker points (added to distance measurements)
        dim : int, optional
            Dimension of ambient space (default 3)
        X : ndarray, optional
            Array of column vectors with locations of markers
        """

        # set the marker diameter
        self.diameter = diameter

        if EDM is not None:
            self.dim = dim
            self.fromEDM(EDM, labels=labels)

        elif X is not None:
            self.m = X.shape[1]
            self.dim = X.shape[0]
            self.X = X

            if labels is None:
                self.labels = [str(i) for i in range(self.m)]

        else:
            self.m = m
            self.dim = dim
            self.X = np.zeros((self.dim, self.m))

        # Now set the labels
        if labels is not None:
            if len(labels) == self.m:
                self.labels = labels
            else:
                raise ValueError("There needs to be one label per marker point")
        else:
            self.labels = [str(i) for i in range(self.m)]

    def __getitem__(self, ref):

        if isinstance(ref, (str, unicode)):
            if self.labels is None:
                raise ValueError("Labels not set for this marker set")
            index = self.labels.index(ref)
        elif isinstance(ref, int) or isinstance(ref, slice):
            index = ref
        elif isinstance(ref, list):
            index = [
                self.labels.index(s) if isinstance(s, (str, unicode)) else s
                for s in ref
            ]
        else:
            index = int(ref)

        return self.X[:, index]

    def copy(self):
        """Return a deep copy of this marker set object"""

        new_marker = PointCloud(
            X=self.X.copy(), labels=self.labels, diameter=self.diameter
        )
        return new_marker

    def key2ind(self, ref):
        """Get the index location from a label"""

        if isinstance(ref, (str, unicode)):
            if self.labels is None:
                raise ValueError("Labels must be defined to be used to access markers")
            else:
                return self.labels.index(ref)
        else:
            return int(ref)

    def fromEDM(self, D, labels=None, method="mds"):
        """
        Compute the position of markers from their Euclidean Distance Matrix

        Parameters
        ----------
        D: square 2D ndarray
            Euclidean Distance Matrix (matrix containing squared distances between points
        labels: list, optional
            A list of human friendly labels for the markers (e.g. 'east', 'west', etc)
        method: str, optional
            The method to use
            * 'mds' for multidimensional scaling (default)
            * 'tri' for trilateration
        """

        if D.shape[0] != D.shape[1]:
            raise ValueError("The distance matrix must be square")

        self.m = D.shape[0]

        if method == "tri":
            self.trilateration(D)
        else:
            self.classical_mds(D)

    def classical_mds(self, D):
        """
        Classical multidimensional scaling

        Parameters
        ----------
        D : square 2D ndarray
            Euclidean Distance Matrix (matrix containing squared distances between points
        """

        # Apply MDS algorithm for denoising
        n = D.shape[0]
        J = np.eye(n) - np.ones((n, n)) / float(n)
        G = -0.5 * np.dot(J, np.dot(D, J))

        s, U = np.linalg.eig(G)

        # we need to sort the eigenvalues in decreasing order
        s = np.real(s)
        o = np.argsort(s)
        s = s[o[::-1]]
        U = U[:, o[::-1]]

        S = np.diag(s)[0 : self.dim, :]
        self.X = np.dot(np.sqrt(S), U.T)

    def trilateration_single_point(self, c, Dx, Dy):
        """
        Given x at origin (0,0) and y at (0,c) the distances from a point
        at unknown location Dx, Dy to x, y, respectively, finds the position of the point.
        """

        z = (c**2 - (Dy**2 - Dx**2)) / (2 * c)
        t = np.sqrt(Dx**2 - z**2)

        return np.array([t, z])

    def trilateration(self, D):
        """
        Find the location of points based on their distance matrix using trilateration

        Parameters
        ----------
        D : square 2D ndarray
            Euclidean Distance Matrix (matrix containing squared distances between points
        """

        dist = np.sqrt(D)

        # Simpler algorithm (no denoising)
        self.X = np.zeros((self.dim, self.m))

        self.X[:, 1] = np.array([0, dist[0, 1]])
        for i in range(2, m):
            self.X[:, i] = self.trilateration_single_point(
                self.X[1, 1], dist[0, i], dist[1, i]
            )

    def EDM(self):
        """Computes the EDM corresponding to the marker set"""
        if self.X is None:
            raise ValueError("No marker set")

        G = np.dot(self.X.T, self.X)
        return (
            np.outer(np.ones(self.m), np.diag(G))
            - 2 * G
            + np.outer(np.diag(G), np.ones(self.m))
        )

    def normalize(self, refs=None):
        """
        Reposition points such that x0 is at origin, x1 lies on c-axis
        and x2 lies above x-axis, keeping the relative position to each other.
        The z-axis is defined according to right hand rule by default.

        Parameters
        ----------
        refs : list of 3 ints or str
            The index or label of three markers used to define (origin, x-axis, y-axis)
        left_hand : bool, optional (default False)
            Normally the z-axis is defined using right-hand rule, this flag allows to override this behavior
        """

        if refs is None:
            refs = [0, 1, 2, 3]

        # Transform references to indices if needed
        refs = [self.key2ind(s) for s in refs]

        if self.dim == 2 and len(refs) < 3:
            raise ValueError(
                "In 2D three reference points are needed to define a reference frame"
            )
        elif self.dim == 3 and len(refs) < 4:
            raise ValueError(
                "In 3D four reference points are needed to define a reference frame"
            )

        # set first point to origin
        X0 = self.X[:, refs[0], None]
        Y = self.X - X0

        # Rotate around z to align x-axis to second point
        theta = np.arctan2(Y[1, refs[1]], Y[0, refs[1]])
        c = np.cos(theta)
        s = np.sin(theta)
        if self.dim == 2:
            H = np.array([[c, s], [-s, c]])
        elif self.dim == 3:
            H = np.array([[c, s, 0], [-s, c, 0], [0, 0, 1]])
        Y = np.dot(H, Y)

        if self.dim == 2:
            # set third point to lie above x-axis
            if Y[1, refs[2]] < 0:
                Y[1, :] *= -1

        elif self.dim == 3:
            # In 3D we also want to make sur z-axis points up
            theta = np.arctan2(Y[2, refs[2]], Y[1, refs[2]])
            c = np.cos(theta)
            s = np.sin(theta)
            H = np.array([[1, 0, 0], [0, c, s], [0, -s, c]])
            Y = np.dot(H, Y)

        # Flip the z-axis if requested
        if self.dim == 3 and Y[2, refs[3]] < 0:
            Y[2, :] *= -1

        self.X = Y

    def center(self, marker):
        """Translate the marker set so that the argument is the origin."""

        index = self.key2ind(marker)
        self.X -= self.X[:, index, None]

    def align(self, marker, axis):
        """
        Rotate the marker set around the given axis until it is aligned onto the given marker

        Parameters
        ----------
        marker : int or str
            the index or label of the marker onto which to align the set
        axis : int
            the axis around which the rotation happens
        """

        index = self.key2ind(marker)
        axis = (
            ["x", "y", "z"].index(axis) if isinstance(marker, (str, unicode)) else axis
        )

        # swap the axis around which to rotate to last position
        Y = self.X
        if self.dim == 3:
            Y[axis, :], Y[2, :] = Y[2, :], Y[axis, :]

        # Rotate around z to align x-axis to second point
        theta = np.arctan2(Y[1, index], Y[0, index])
        c = np.cos(theta)
        s = np.sin(theta)
        H = np.array([[c, s], [-s, c]])
        Y[:2, :] = np.dot(H, Y[:2, :])

        if self.dim == 3:
            Y[axis, :], Y[2, :] = Y[2, :], Y[axis, :]

    def flatten(self, ind):
        """
        Transform the set of points so that the subset of markers given as argument is
        as close as flat (wrt z-axis) as possible.

        Parameters
        ----------
        ind : list of bools
            Lists of marker indices that should be all in the same subspace
        """

        # Transform references to indices if needed
        ind = [self.key2ind(s) for s in ind]

        # center point cloud around the group of indices
        centroid = self.X[:, ind].mean(axis=1, keepdims=True)
        X_centered = self.X - centroid

        # The rotation is given by left matrix of SVD
        U, S, V = la.svd(X_centered[:, ind], full_matrices=False)

        self.X = np.dot(U.T, X_centered) + centroid

    def correct(self, corr_dic):
        """correct a marker location by a given vector"""

        for key, val in corr_dic.items():
            ind = self.key2ind(key)
            self.X[:, ind] += val

    def doa(self, receiver, source):
        """Computes the direction of arrival wrt a source and receiver"""

        s_ind = self.key2ind(source)
        r_ind = self.key2ind(receiver)

        # vector from receiver to source
        v = self.X[:, s_ind] - self.X[:, r_ind]

        azimuth = np.arctan2(v[1], v[0])
        elevation = np.arctan2(v[2], la.norm(v[:2]))

        return np.array([azimuth, elevation])

    def plot(self, axes=None, show_labels=True, **kwargs):

        try:
            from mpl_toolkits.mplot3d import Axes3D
            import matplotlib.pyplot as plt
        except ImportError:
            import warnings

            warnings.warn("Matplotlib is required for plotting")
            return

        if self.dim == 2:

            # Create a figure if needed
            if axes is None:
                axes = plt.subplot(111)

            axes.plot(self.X[0, :], self.X[1, :], **kwargs)
            axes.axis(aspect="equal")
            plt.show()

        elif self.dim == 3:
            if axes is None:
                fig = plt.figure()
                axes = fig.add_subplot(111, projection="3d")
            axes.scatter(self.X[0, :], self.X[1, :], self.X[2, :], **kwargs)

            axes.set_xlabel("X")
            axes.set_ylabel("Y")
            axes.set_zlabel("Z")
            plt.show()

        if show_labels and self.labels is not None:
            eps = np.linalg.norm(self.X[:, 0] - self.X[:, 1]) / 100
            for i in range(self.m):
                if self.dim == 2:
                    axes.text(self.X[0, i] + eps, self.X[1, i] + eps, self.labels[i])
                elif self.dim == 3:
                    axes.text(
                        self.X[0, i] + eps,
                        self.X[1, i] + eps,
                        self.X[2, i] + eps,
                        self.labels[i],
                        None,
                    )

        return axes


if __name__ == "__main__":

    import matplotlib.pyplot as plt

    # number of markers
    m = 4
    dim = 2

    D = np.zeros((m, m))

    marker_diameter = 0.040  # 4 cm

    M_orig = MarkerSet(X=np.array([[0.0, 0.0], [0.7, 0.0], [0.7, 0.7], [0.0, 0.7]]).T)
    D = np.sqrt(M_orig.EDM())

    """
    D[0,1] = D[1,0] = 4.126 + marker_diameter
    D[0,2] = D[2,0] = 6.878 + marker_diameter
    D[0,3] = D[3,0] = 4.508 + marker_diameter
    D[1,2] = D[2,1] = 4.401 + marker_diameter
    D[1,3] = D[3,1] = 7.113 + marker_diameter
    D[3,2] = D[2,3] = 7.002 + marker_diameter
    """

    M1 = MarkerSet(m=m, dim=dim, diameter=marker_diameter)
    M1.fromEDM(D**2)
    M1.normalize()

    M2 = MarkerSet(m=m, dim=dim, diameter=marker_diameter)
    M2.fromEDM(D**2, method="tri")
    M2.normalize()

    M2.plot(marker="ko", labels=True)
    M1.plot(marker="rx")
    plt.show()



from __future__ import division, print_function

import numpy as np
import pyroomacoustics as pra
from pyroomacoustics.transform import STFT
import time
import warnings

# test signal
np.random.seed(0)
num_mic = 25
signals = np.random.randn(100000, num_mic).astype(np.float32)
fs = 16000

# STFT parameters
block_size = 512
hop = block_size // 2
win = pra.hann(block_size)
x_r = np.zeros(signals.shape)
num_times = 50

print()

"""
One frame at a time
"""
print(
    "Averaging computation time over %d cases of %d channels of %d samples (%0.1f s at %0.1f kHz)."
    % (num_times, num_mic, len(signals), (len(signals) / fs), fs / 1000)
)
print()
print("----- SINGLE FRAME AT A TIME -----")
print("With STFT object (not fixed) : ", end="")
stft = STFT(block_size, hop=hop, channels=num_mic, streaming=True, analysis_window=win)
start = time.time()
for k in range(num_times):

    x_r = np.zeros(signals.shape)
    n = 0
    while signals.shape[0] - n > hop:
        stft.analysis(
            signals[
                n : n + hop,
            ]
        )
        x_r[
            n : n + hop,
        ] = stft.synthesis()
        n += hop
avg_time = (time.time() - start) / num_times
print("%0.3f sec" % avg_time)
err_dB = 20 * np.log10(
    np.max(
        np.abs(
            signals[
                : n - hop,
            ]
            - x_r[
                hop:n,
            ]
        )
    )
)
print("Error [dB] : %0.3f" % err_dB)


print("With STFT object (fixed) : ", end="")
stft = STFT(
    block_size,
    hop=hop,
    channels=num_mic,
    num_frames=1,
    streaming=True,
    analysis_window=win,
)
start = time.time()
for k in range(num_times):

    x_r = np.zeros(signals.shape)
    n = 0
    while signals.shape[0] - n > hop:
        stft.analysis(
            signals[
                n : n + hop,
            ]
        )
        x_r[
            n : n + hop,
        ] = stft.synthesis()
        n += hop
avg_time = (time.time() - start) / num_times
print("%0.3f sec" % avg_time)
err_dB = 20 * np.log10(
    np.max(
        np.abs(
            signals[
                : n - hop,
            ]
            - x_r[
                hop:n,
            ]
        )
    )
)
print("Error [dB] : %0.3f" % err_dB)


"""
Multiple frame at a time (non-streaming)
"""
print()
print("----- MULTIPLE FRAMES AT A TIME -----")

warnings.filterwarnings("ignore")  # to avoid warning of appending zeros to be printed
print("With STFT object (not fixed) : ", end="")
stft = STFT(block_size, hop=hop, channels=num_mic, analysis_window=win, streaming=False)
start = time.time()
for k in range(num_times):

    stft.analysis(signals)
    x_r = stft.synthesis()

avg_time = (time.time() - start) / num_times
print("%0.3f sec" % avg_time)
err_dB = 20 * np.log10(
    np.max(np.abs(signals[hop : len(x_r) - hop] - x_r[hop : len(x_r) - hop]))
)
print("Error [dB] : %0.3f" % err_dB)
warnings.filterwarnings("default")


print("With STFT object (fixed) : ", end="")
num_frames = (len(signals) - block_size) // hop + 1
stft = STFT(
    block_size,
    hop=hop,
    channels=num_mic,
    num_frames=num_frames,
    analysis_window=win,
    streaming=False,
)
start = time.time()
for k in range(num_times):

    stft.analysis(signals[: (num_frames - 1) * hop + block_size, :])
    x_r = stft.synthesis()
avg_time = (time.time() - start) / num_times
print("%0.3f sec" % avg_time)
err_dB = 20 * np.log10(
    np.max(np.abs(signals[hop : len(x_r) - hop] - x_r[hop : len(x_r) - hop]))
)
print("Error [dB] : %0.3f" % err_dB)



"""
Block based processing
======================

This subpackage contains routines for continuous *realtime-like* block
processing of signals with the STFT. The routines are written to take
advantage of fast FFT libraries like `pyfftw` or `mkl` when available.

DFT
    | A class for performing DFT of real signals
    | :py:obj:`pyroomacoustics.transform.dft`
STFT
    | A class for continuous STFT processing and frequency domain filtering
    | :py:obj:`pyroomacoustics.transform.stft`

"""


from . import stft
from .dft import DFT
from .stft import STFT



import numpy as np
from pyroomacoustics import angle_function
from unittest import TestCase


pi = np.pi

# for 3-D coordinates
a1 = np.array([[0, 0, 1], [0, 0, 1], [0, 1, 1]])
a2 = np.array([[0], [0], [1]])
a3 = np.array([0, 0, 1]).T

b1 = np.array([[0], [0], [0]])
b2 = np.array([1, -1, 1]).T


a1_b1 = np.array([[0, 0, pi / 4], [0, 0, pi / 4]])
a1_b2 = np.array([[3 * pi / 4, 3 * pi / 4, pi / 2], [3 * pi / 4, pi / 2, pi / 2]])
a2_b1 = np.array([[0], [0]])
a2_b2 = np.array([[3 * pi / 4], [pi / 2]])
a3_b1 = np.array([[0], [0]])
a3_b2 = np.array([[3 * pi / 4], [pi / 2]])


# for 2-D coordinates
c1 = np.array([[0, 0], [0, 0]])
c2 = np.array([[1], [1]])
c3 = np.array([1, 1]).T

d1 = np.array([[0], [0]])
d2 = np.array([0, 1]).T


c1_d1 = np.array([[0, 0], [pi / 2, pi / 2]])
c1_d2 = np.array([[-pi / 2, -pi / 2], [pi / 2, pi / 2]])
c2_d1 = np.array([[pi / 4], [pi / 2]])
c2_d2 = np.array([[0], [pi / 2]])
c3_d1 = np.array([[pi / 4], [pi / 2]])
c3_d2 = np.array([[0], [pi / 2]])


class TestAngleFunction(TestCase):
    def test_set_3d(self):
        self.assertTrue(angle_function(a1, b1).all() == a1_b1.all())
        self.assertTrue(angle_function(a1, b2).all() == a1_b2.all())

    def test_point_3d_1(self):
        self.assertTrue(angle_function(a2, b1).all() == a2_b1.all())
        self.assertTrue(angle_function(a2, b2).all() == a2_b2.all())

    def test_point_3d_2(self):
        self.assertTrue(angle_function(a3, b1).all() == a3_b1.all())
        self.assertTrue(angle_function(a3, b2).all() == a3_b2.all())

    def test_set_2d(self):
        self.assertTrue(angle_function(c1, d1).all() == c1_d1.all())
        self.assertTrue(angle_function(c1, d2).all() == c1_d2.all())

    def test_point_2d_1(self):
        self.assertTrue(angle_function(c2, d1).all() == c2_d1.all())
        self.assertTrue(angle_function(c2, d2).all() == c2_d2.all())

    def test_point_2d_2(self):
        self.assertTrue(angle_function(c3, d1).all() == c3_d1.all())
        self.assertTrue(angle_function(c3, d2).all() == c3_d2.all())


def find_error(type_coordinates):
    if type_coordinates == "3-D":
        print("-" * 40)
        print("type_coordinates = 3-D")
        print("-" * 40)

        a_range = [a1, a2, a3]
        b_range = [b1, b2]
        a_b_range = [a1_b1, a1_b2, a2_b1, a2_b2, a3_b1, a3_b2]

        a_b_index = 0
        for a in a_range:
            for b in b_range:

                error_azimuth = (angle_function(a, b) - a_b_range[a_b_index])[0]
                error_colatitude = (angle_function(a, b) - a_b_range[a_b_index])[1]

                print("for points :")
                print(a)
                print(b)
                print(
                    "error in azimuth calculation: {}".format(np.average(error_azimuth))
                )
                print(
                    "error in colatitude calculation: {}".format(
                        np.average(error_colatitude)
                    )
                )
                print()
                a_b_index += 1

    elif type_coordinates == "2-D":
        print("-" * 40)
        print("type_coordinates = 2-D")
        print("-" * 40)

        c_range = [c1, c2, c3]
        d_range = [d1, d2]
        c_d_range = [c1_d1, c1_d2, c2_d1, c2_d2, c3_d1, c3_d2]

        c_d_index = 0
        for c in c_range:
            for d in d_range:

                error_azimuth = (angle_function(c, d) - c_d_range[c_d_index])[0]
                error_colatitude = (angle_function(c, d) - c_d_range[c_d_index])[1]

                print("for points :")
                print(c)
                print(d)
                print(
                    "error in azimuth calculation: {}".format(np.average(error_azimuth))
                )
                print(
                    "error in colatitude calculation: {}".format(
                        np.average(error_colatitude)
                    )
                )
                print()
                c_d_index += 1


if __name__ == "__main__":
    find_error("3-D")
    find_error("2-D")
